import argparse
import logging

import numpy as np
import xarray as xr

from icenet.data.cli import date_arg
from icenet.utils import setup_logging


def threshold_exceeds(da: object,
                      sic_thresh: float,
                      window_length: int = 1,
                      dimensions: dict = None):
    """

    :param da:
    :param sic_thresh:
    :param window_length:
    :param dimensions:
    :return:
    """
    logging.info("Checking thresholds for forecast(s)")

    if dimensions:
        logging.debug("Selecting within given dimensions: {}".
                      format(dimensions))
        da = da.sel(**dimensions)

    thresh_arr = da > sic_thresh

    window_acc = thresh_arr.rolling(leadtime=window_length).reduce(np.sum)
    threshold_exceed_arr = xr.where(window_acc == window_length, da, 0)

    return np.argwhere(threshold_exceed_arr.values)


@setup_logging
def threshold_args():
    ap = argparse.ArgumentParser()
    ap.add_argument("-s", "--date-start", type=date_arg, default=None)
    ap.add_argument("-e", "--date-end", type=date_arg, default=None)

    ap.add_argument("-o", "--output-file", default=None, type=str)
    ap.add_argument("-v", "--verbose", action="store_true", default=False)

    ap.add_argument("forecast_file")
    ap.add_argument("threshold", type=float)
    ap.add_argument("window_length", type=int)
    args = ap.parse_args()
    return args


def threshold_main():
    args = threshold_args()
    da = xr.open_dataset(args.forecast_file).sic_mean
    threshold_map = threshold_exceeds(da, args.threshold, args.window_length)

    # Currently naive implementation
    if args.output_file:
        with open(args.output_file, "wb") as fh:
            np.save(fh, threshold_map)
        logging.info("Saved to {}".format(args.output_file))
    else:
        logging.info("No output file provided: {} cells breached threshold".
                     format(len(threshold_map)))



import concurrent
import logging
import os
import re
import shutil
import tempfile

from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from itertools import product

from icenet.data.sic.mask import Masks
from icenet.data.sic.utils import SIC_HEMI_STR
from icenet.data.producers import Downloader
from icenet.data.utils import assign_lat_lon_coord_system, \
    gridcell_angles_from_dim_coords, \
    invert_gridcell_angles, \
    rotate_grid_vectors
from icenet.data.interfaces.utils import batch_requested_dates
from icenet.utils import run_command

import iris
import iris.analysis
import iris.cube
import iris.exceptions
import numpy as np
import pandas as pd
import xarray as xr

"""

"""


def filter_dates_on_data(latlon_path: str,
                         regridded_name: str,
                         req_dates: object,
                         check_latlon: bool = True,
                         check_regridded: bool = True,
                         drop_vars: list = None):
    """Reduces request dates and target files based on existing data

    To avoid what is potentially significant resource expense downloading
    extant data, downloaders should call this method to reduce the request
    dates only to that data not already present. This is a fairly naive
    implementation, in that if the data is present in either the latlon
    intermediate file OR the target regridded file, we'll not bother
    downloading again. This can be overridden via the method arguments.

    :param latlon_path:
    :param regridded_name:
    :param req_dates:
    :param check_latlon:
    :param check_regridded:
    :param drop_vars:
    :return: req_dates(list)
    """

    latlon_dates = list()
    regridded_dates = list()
    drop_vars = list() if drop_vars is None else drop_vars

    # Latlon files should in theory be aggregated and singular arrays
    # meaning we can naively open and interrogate the dates
    if check_latlon and os.path.exists(latlon_path):
        try:
            latlon_dates = xr.open_dataset(
                latlon_path,
                drop_variables=drop_vars).time.values
            logging.debug("{} latlon dates already available in {}".format(
                len(latlon_dates), latlon_path
            ))
        except ValueError:
            logging.warning("Latlon {} dates not readable, ignoring file")

    if check_regridded and os.path.exists(regridded_name):
        regridded_dates = xr.open_dataset(
            regridded_name,
            drop_variables=drop_vars).time.values
        logging.debug("{} regridded dates already available in {}".format(
            len(regridded_dates), regridded_name
        ))

    exclude_dates = list(set(latlon_dates).union(set(regridded_dates)))
    logging.debug("Excluding {} dates already existing from {} dates "
                  "requested.".format(len(exclude_dates), len(req_dates)))

    return sorted(list(pd.to_datetime(req_dates).
                       difference(pd.to_datetime(exclude_dates))))


def merge_files(new_datafile: str,
                other_datafile: str,
                drop_variables: object = None):
    """

    :param new_datafile:
    :param other_datafile:
    :param drop_variables:
    """
    drop_variables = list() if drop_variables is None else drop_variables

    if other_datafile is not None:
        (datafile_path, new_filename) = os.path.split(new_datafile)
        moved_new_datafile = \
            os.path.join(datafile_path, "new.{}".format(new_filename))
        os.rename(new_datafile, moved_new_datafile)
        d1 = xr.open_dataarray(moved_new_datafile,
                               drop_variables=drop_variables)

        logging.info("Concatenating with previous data {}".format(
            other_datafile
        ))
        d2 = xr.open_dataarray(other_datafile,
                               drop_variables=drop_variables)
        new_ds = xr.concat([d1, d2], dim="time").\
            sortby("time").\
            drop_duplicates("time", keep="first")

        logging.info("Saving merged data to {}... ".
                     format(new_datafile))
        new_ds.to_netcdf(new_datafile)
        os.unlink(other_datafile)
        os.unlink(moved_new_datafile)


class ClimateDownloader(Downloader):
    """Climate downloader base class

    :param dates:
    :param delete_tempfiles:
    :param download:
    :param group_dates_by:
    :param max_threads:
    :param postprocess:
    :param pregrid_prefix:
    :param levels:
    :param var_name_idx:
    :param var_names:
    """

    def __init__(self, *args,
                 dates: object = (),
                 delete_tempfiles: bool = True,
                 download: bool = True,
                 drop_vars: list = None,
                 group_dates_by: str = "year",
                 levels: object = (),
                 max_threads: int = 1,
                 postprocess: bool = True,
                 pregrid_prefix: str = "latlon_",
                 var_name_idx: int = -1,
                 var_names: object = (),
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._dates = list(dates)
        self._delete = delete_tempfiles
        self._download = download
        self._drop_vars = list() if drop_vars is None else drop_vars
        self._files_downloaded = []
        self._group_dates_by = group_dates_by
        self._levels = list(levels)
        self._masks = Masks(north=self.north, south=self.south)
        self._max_threads = max_threads
        self._postprocess = postprocess
        self._pregrid_prefix = pregrid_prefix
        self._rotatable_files = []
        self._sic_ease_cubes = dict()
        self._var_name_idx = var_name_idx
        self._var_names = list(var_names)

        assert len(self._var_names), "No variables requested"
        assert len(self._levels) == len(self._var_names), \
            "# of levels must match # vars"

        if not self._delete:
            logging.warning("!!! Deletions of temp files are switched off: be "
                            "careful with this, you need to manage your "
                            "files manually")
        self._download_method = None

        self._validate_config()

    def _validate_config(self):
        """

        """
        if self.hemisphere_str in os.path.split(self.base_path):
            raise RuntimeError("Don't include hemisphere string {} in "
                               "base path".format(self.hemisphere_str))

    def download(self):
        """Handles concurrent (threaded) downloading for variables

        This takes dates, variables and levels as configured, batches them into
        requests and submits those via a ThreadPoolExecutor for concurrent
        downloading. Returns nothing, relies on _single_download to implement
        appropriate updates to this object to record state changes arising from
        downloading.
        """

        logging.info("Building request(s), downloading and daily averaging "
                     "from {} API".format(self.identifier.upper()))

        requests = list()

        for idx, var_name in enumerate(self.var_names):
            levels = [None] if not self.levels[idx] else self.levels[idx]

            dates_per_request = \
                batch_requested_dates(self._dates,
                                      attribute=self._group_dates_by)

            for var_prefix, level, req_date in \
                    product([var_name], levels, dates_per_request):
                requests.append((var_prefix, level, req_date))

        with ThreadPoolExecutor(max_workers=min(len(requests),
                                                self._max_threads)) \
                as executor:
            futures = []

            for var_prefix, level, req_date in requests:
                future = executor.submit(self._single_download,
                                         var_prefix,
                                         level,
                                         req_date)
                futures.append(future)

            for future in concurrent.futures.as_completed(futures):
                try:
                    future.result()
                except Exception as e:
                    logging.exception("Thread failure: {}".format(e))

        logging.info("{} daily files downloaded".
                     format(len(self._files_downloaded)))

    def _single_download(self,
                         var_prefix: str,
                         level: object,
                         req_dates: object):
        """Implements a single download based on configured download_method

        This allows delegation of downloading logic in a consistent manner to
        the configured download_method, ensuring a guarantee of adherence to
        naming and processing flow within ClimateDownloader implementations.

        :param var_prefix: the icenet variable name
        :param level: the height to download
        :param req_dates: the request date
        """

        logging.info("Processing single download for {} @ {} with {} dates".
                     format(var_prefix, level, len(req_dates)))
        var = var_prefix if not level else \
            "{}{}".format(var_prefix, level)
        var_folder = self.get_data_var_folder(var)

        latlon_path, regridded_name = \
            self.get_req_filenames(var_folder, req_dates[0])

        req_dates = filter_dates_on_data(latlon_path,
                                         regridded_name,
                                         req_dates,
                                         drop_vars=self._drop_vars)

        if len(req_dates):
            if self._download:
                with tempfile.TemporaryDirectory() as tmpdir:
                    tmp_latlon_path = os.path.join(tmpdir, os.path.basename("{}.download".format(latlon_path)))

                    self.download_method(var,
                                         level,
                                         req_dates,
                                         tmp_latlon_path)

                    if os.path.exists(latlon_path):
                        (ll_path, ll_file) = os.path.split(latlon_path)
                        rename_latlon_path = os.path.join(
                            ll_path, "{}_old{}".format(
                                *os.path.splitext(ll_file)))
                        os.rename(latlon_path, rename_latlon_path)
                        old_da = xr.open_dataarray(rename_latlon_path,
                                                   drop_variables=self._drop_vars)
                        tmp_da = xr.open_dataarray(tmp_latlon_path,
                                                   drop_variables=self._drop_vars)

                        logging.debug("Input (old): \n{}".format(old_da))
                        logging.debug("Input (dl): \n{}".format(tmp_da))

                        da = xr.concat([old_da, tmp_da], dim="time")
                        logging.debug("Output: \n{}".format(da))

                        da.to_netcdf(latlon_path)
                        old_da.close()
                        tmp_da.close()
                        os.unlink(rename_latlon_path)
                    else:
                        shutil.move(tmp_latlon_path, latlon_path)

                logging.info("Downloaded to {}".format(latlon_path))
            else:
                logging.info("Skipping actual download to {}".
                             format(latlon_path))
        else:
            logging.info("No requested dates remain, likely already present")

        if self._postprocess and os.path.exists(latlon_path):
            self.postprocess(var, latlon_path)

        if os.path.exists(latlon_path):
            self._files_downloaded.append(latlon_path)

    def postprocess(self, var, download_path):
        logging.debug("No postprocessing in place for {}: {}".
                      format(var, download_path))

    def save_temporal_files(self, var, da,
                            date_format=None,
                            freq=None):
        """

        :param var:
        :param da:
        :param date_format:
        :param freq:
        """
        var_folder = self.get_data_var_folder(var)
        group_by = "time.{}".format(self._group_dates_by) if not freq else freq

        for dt, dt_da in da.groupby(group_by):
            req_date = pd.to_datetime(dt_da.time.values[0])
            latlon_path, regridded_name = \
                self.get_req_filenames(var_folder,
                                       req_date,
                                       date_format=date_format)

            logging.info("Retrieving and saving {}".format(latlon_path))
            dt_da.compute()
            dt_da.to_netcdf(latlon_path)

            if not os.path.exists(regridded_name):
                self._files_downloaded.append(latlon_path)

    @property
    def sic_ease_cube(self):
        """

        :return sic_cube:
        """
        if self._hemisphere not in self._sic_ease_cubes:
            sic_day_fname = 'ice_conc_{}_ease2-250_cdr-v2p0_197901021200.nc'. \
                format(SIC_HEMI_STR[self.hemisphere_str[0]])
            sic_day_path = os.path.join(self.get_data_var_folder("siconca"),
                                        sic_day_fname)
            if not os.path.exists(sic_day_path):
                logging.info("Downloading single daily SIC netCDF file for "
                             "regridding ERA5 data to EASE grid...")

                retrieve_sic_day_cmd = 'wget -m -nH --cut-dirs=6 -P {} ' \
                                       'ftp://osisaf.met.no/reprocessed/ice/' \
                                       'conc/v2p0/1979/01/{}'.\
                    format(self.get_data_var_folder("siconca"), sic_day_fname)

                run_command(retrieve_sic_day_cmd)

            # Load a single SIC map to obtain the EASE grid for
            # regridding ERA data
            self._sic_ease_cubes[self._hemisphere] = \
                iris.load_cube(sic_day_path, 'sea_ice_area_fraction')

            # Convert EASE coord units to metres for regridding
            self._sic_ease_cubes[self._hemisphere].coord(
                'projection_x_coordinate').convert_units('meters')
            self._sic_ease_cubes[self._hemisphere].coord(
                'projection_y_coordinate').convert_units('meters')
        return self._sic_ease_cubes[self._hemisphere]

    def regrid(self,
               files: object = None,
               rotate_wind: bool = True):
        """

        :param files:
        """
        filelist = self._files_downloaded if not files else files
        batches = [filelist[b:b + 1000] for b in range(0, len(filelist), 1000)]

        max_workers = min(len(batches), self._max_threads)
        regrid_results = list()

        if max_workers > 0:
            with ThreadPoolExecutor(max_workers=max_workers) \
                    as executor:
                futures = []

                for files in batches:
                    future = executor.submit(self._batch_regrid, files)
                    futures.append(future)

                for future in concurrent.futures.as_completed(futures):
                    try:
                        fut_results = future.result()

                        for res in fut_results:
                            logging.debug("Future result -> regrid_results: {}".
                                          format(res))
                            regrid_results.append(res)
                    except Exception as e:
                        logging.exception("Thread failure: {}".format(e))
        else:
            logging.info("No regrid batches to processing, moving on...")

        if rotate_wind:
            logging.info("Rotating wind data prior to merging")
            self.rotate_wind_data()

        for new_datafile, moved_datafile in regrid_results:
            merge_files(new_datafile, moved_datafile, self._drop_vars)

    def _batch_regrid(self,
                      files: object):
        """

        :param files:
        """
        results = list()

        for datafile in files:
            (datafile_path, datafile_name) = os.path.split(datafile)

            new_filename = re.sub(r'^{}'.format(
                self.pregrid_prefix), '', datafile_name)
            new_datafile = os.path.join(datafile_path, new_filename)

            moved_datafile = None

            if os.path.exists(new_datafile):
                moved_filename = "moved.{}".format(new_filename)
                moved_datafile = os.path.join(datafile_path, moved_filename)
                os.rename(new_datafile, moved_datafile)

                logging.info("{} already existed, moved to {}".
                             format(new_filename, moved_filename))

            logging.debug("Regridding {}".format(datafile))

            try:
                cube = iris.load_cube(datafile)
                cube = self.convert_cube(cube)

                cube_ease = cube.regrid(
                    self.sic_ease_cube, iris.analysis.Linear())

            except iris.exceptions.CoordinateNotFoundError:
                logging.warning("{} has no coordinates...".
                                format(datafile_name))
                if self.delete:
                    logging.debug("Deleting failed file {}...".
                                  format(datafile_name))
                    os.unlink(datafile)
                continue

            self.additional_regrid_processing(datafile, cube_ease)

            logging.info("Saving regridded data to {}... ".format(new_datafile))
            iris.save(cube_ease, new_datafile, fill_value=np.nan)
            results.append((new_datafile, moved_datafile))

            if self.delete:
                logging.info("Removing {}".format(datafile))
                os.remove(datafile)

        return results

    def convert_cube(self, cube: object):
        """Converts Iris cube to be fit for regrid

        :param cube: the cube requiring alteration
        :return cube: the altered cube
        """

        cube = assign_lat_lon_coord_system(cube)
        return cube

    @abstractmethod
    def additional_regrid_processing(self,
                                     datafile: str,
                                     cube_ease: object):
        """

        :param datafile:
        :param cube_ease:
        """
        pass

    def rotate_wind_data(self,
                         apply_to: object = ("uas", "vas"),
                         manual_files: object = None):
        """

        :param apply_to:
        :param manual_files:
        """
        assert len(apply_to) == 2, "Too many wind variables supplied: {}, " \
                                   "there should only be two.".\
            format(", ".join(apply_to))

        angles = gridcell_angles_from_dim_coords(self.sic_ease_cube)
        invert_gridcell_angles(angles)

        logging.info("Rotating wind data in {}".format(
            " ".join([self.get_data_var_folder(v) for v in apply_to])))

        wind_files = {}

        for var in apply_to:
            source = self.get_data_var_folder(var)

            file_source = self._files_downloaded \
                if not manual_files else manual_files

            latlon_files = [df for df in file_source if source in df]
            wind_files[var] = sorted([
                re.sub(r'{}'.format(self.pregrid_prefix), '', df)
                for df in latlon_files
                if os.path.dirname(df).split(os.sep)
                [self._var_name_idx] == var],
                key=lambda x: int(re.search(r'^(?:\w+_)?(\d+).nc',
                                  os.path.basename(x)).group(1))
            )
            logging.info("{} files for {}".format(len(wind_files[var]), var))

        # NOTE: we're relying on apply_to having equal datasets
        assert len(wind_files[apply_to[0]]) == len(wind_files[apply_to[1]]), \
            "The wind file datasets are unequal in length"

        # validation
        for idx, wind_file_0 in enumerate(wind_files[apply_to[0]]):
            wind_file_1 = wind_files[apply_to[1]][idx]

            wd0 = re.sub(r'^{}_'.format(apply_to[0]), '',
                         os.path.basename(wind_file_0))

            if not wind_file_1.endswith(wd0):
                logging.error("Wind file array is not valid: {}".format(
                    zip(wind_files)))
                raise RuntimeError("{} is not at the end of {}, something is "
                                   "wrong".format(wd0, wind_file_1))

        for idx, wind_file_0 in enumerate(wind_files[apply_to[0]]):
            wind_file_1 = wind_files[apply_to[1]][idx]

            logging.info("Rotating {} and {}".format(wind_file_0, wind_file_1))

            wind_cubes = dict()
            wind_cubes_r = dict()

            wind_cubes[apply_to[0]] = iris.load_cube(wind_file_0)
            wind_cubes[apply_to[1]] = iris.load_cube(wind_file_1)

            try:
                wind_cubes_r[apply_to[0]], wind_cubes_r[apply_to[1]] = \
                    rotate_grid_vectors(
                        wind_cubes[apply_to[0]],
                        wind_cubes[apply_to[1]],
                        angles,
                    )
            except iris.exceptions.CoordinateNotFoundError:
                logging.exception("Failure to rotate due to coordinate issues. "
                                  "moving onto next file")
                continue

            # Original implementation is in danger of lost updates
            # due to potential lazy loading
            for i, name in enumerate([wind_file_0, wind_file_1]):
                # NOTE: implementation with tempfile caused problems on NFS
                # mounted filesystem, so avoiding in place of letting iris do it
                temp_name = os.path.join(os.path.split(name)[0],
                                         "temp.{}".format(
                                             os.path.basename(name)))
                logging.debug("Writing {}".format(temp_name))

                iris.save(wind_cubes_r[apply_to[i]], temp_name)
                os.replace(temp_name, name)
                logging.debug("Overwritten {}".format(name))

    def get_req_filenames(self,
                          var_folder: str,
                          req_date: object,
                          date_format: str = None):
        """

        :param var_folder:
        :param req_date:
        :param date_format:
        :return:
        """

        filename_date = getattr(req_date, self._group_dates_by) \
            if not date_format else req_date.strftime(date_format)

        latlon_path = os.path.join(
            var_folder, "{}{}.nc".format(self.pregrid_prefix, filename_date))
        regridded_name = os.path.join(
            var_folder, "{}.nc".format(filename_date))

        logging.debug("Got {} filenames: {} and {}".format(
            self._group_dates_by, latlon_path, regridded_name
        ))

        return latlon_path, regridded_name

    @property
    def dates(self):
        return self._dates

    @property
    def delete(self):
        return self._delete

    @property
    def download_method(self) -> callable:
        if not self._download_method:
            raise RuntimeError("Downloader has no method set, "
                               "implementation error")
        return self._download_method

    @download_method.setter
    def download_method(self, method: callable):
        self._download_method = method

    @property
    def group_dates_by(self):
        return self._group_dates_by

    @property
    def levels(self):
        return self._levels

    @property
    def pregrid_prefix(self):
        return self._pregrid_prefix

    @property
    def var_names(self):
        return self._var_names




import argparse
import datetime as dt
import logging
import os
import sys

from datetime import timedelta

import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.animation import FuncAnimation
from matplotlib.backends.backend_pdf import PdfPages

import seaborn as sns

import numpy as np
import pandas as pd
import dask.array as da
import xarray as xr

from icenet import __version__ as icenet_version
from icenet.data.cli import date_arg
from icenet.data.sic.mask import Masks
from icenet.plotting.utils import (
    filter_ds_by_obs,
    get_forecast_ds,
    get_obs_da,
    get_seas_forecast_da,
    get_seas_forecast_init_dates,
    show_img,
    get_plot_axes,
    process_probes,
    process_regions
)
from icenet.plotting.video import xarray_to_video


def parse_location_or_region(argument: str):
    separator = ','
    # Allow ValueError to propagate if not given sequence of integers
    return tuple(int(s) for s in argument.split(separator))


def location_arg(argument: str):
    try:
        x, y = parse_location_or_region(argument)
        return (x, y)
    except ValueError:
        argparse.ArgumentTypeError(
            "Expected a location (pair of integers separated by a comma)"
        )


def region_arg(argument: str):
    """type handler for region arguments with argparse

    :param argument:

    :return:
    """
    try:
        x1, y1, x2, y2 = parse_location_or_region(argument)

        assert x2 > x1 and y2 > y1, "Region is not valid"
        return x1, y1, x2, y2
    except TypeError:
        raise argparse.ArgumentTypeError(
            "Region argument must be list of four integers")


def compute_binary_accuracy(masks: object,
                            fc_da: object,
                            obs_da: object,
                            threshold: float) -> object:
    """
    Compute the binary class accuracy of a forecast,
    where we consider a binary class prediction of ice with SIC > 15%.
    In particular, we compute the mean percentage of correct
    classifications over the active grid cell area.

    :param masks: an icenet Masks object
    :param fc_da: the forecasts given as an xarray.DataArray object
                  with time, xc, yc coordinates
    :param obs_da: the "ground truth" given as an xarray.DataArray object
                   with time, xc, yc coordinates
    :param threshold: the SIC threshold of interest (in percentage as a fraction),
                      i.e. threshold is between 0 and 1

    :return: binary accuracy for forecast as xarray.DataArray object
    """
    threshold = 0.15 if threshold is None else threshold
    if (threshold < 0) or (threshold > 1):
        raise ValueError("threshold must be a float between 0 and 1")

    # obtain mask
    agcm = masks.get_active_cell_da(obs_da)

    # binary for observed (i.e. truth)
    binary_obs_da = obs_da > threshold

    # binary for forecast
    binary_fc_da = fc_da > threshold

    # compute binary accuracy metric
    binary_fc_da = (binary_fc_da == binary_obs_da). \
        astype(np.float16).weighted(agcm)
    binacc_fc = (binary_fc_da.mean(dim=['yc', 'xc']) * 100)

    return binacc_fc


def plot_binary_accuracy(masks: object,
                         fc_da: object,
                         cmp_da: object,
                         obs_da: object,
                         output_path: object,
                         threshold: float = 0.15) -> object:
    """
    Compute and plot the binary class accuracy of a forecast,
    where we consider a binary class prediction of ice with SIC > 15%.
    In particular, we compute the mean percentage of correct
    classifications over the active grid cell area.
    
    :param masks: an icenet Masks object
    :param fc_da: the forecasts given as an xarray.DataArray object 
                  with time, xc, yc coordinates
    :param cmp_da: a comparison forecast / sea ice data given as an 
                   xarray.DataArray object with time, xc, yc coordinates.
                   If None, will ignore plotting a comparison forecast
    :param obs_da: the "ground truth" given as an xarray.DataArray object
                   with time, xc, yc coordinates
    :param output_path: string specifying the path to store the plot
    :param threshold: the SIC threshold of interest (in percentage as a fraction),
                      i.e. threshold is between 0 and 1
    
    :return: tuple of (binary accuracy for forecast (fc_da), 
                       binary accuracy for comparison (cmp_da))
    """
    binacc_fc = compute_binary_accuracy(masks=masks,
                                        fc_da=fc_da,
                                        obs_da=obs_da,
                                        threshold=threshold)
    fig, ax = plt.subplots(figsize=(12, 6))
    ax.set_title(f"Binary accuracy comparison (threshold SIC = {threshold*100}%)")
    ax.plot(binacc_fc.time, binacc_fc.values, label="IceNet")

    if cmp_da is not None:
        binacc_cmp = compute_binary_accuracy(masks=masks,
                                             fc_da=cmp_da,
                                             obs_da=obs_da,
                                             threshold=threshold)
        ax.plot(binacc_cmp.time, binacc_cmp.values, label="SEAS")
    else:
        binacc_cmp = None

    ax.set_ylabel("Binary accuracy (%)")
    ax.xaxis.set_major_formatter(
        mdates.ConciseDateFormatter(ax.xaxis.get_major_locator()))
    ax.xaxis.set_major_locator(mdates.MonthLocator())
    ax.xaxis.set_minor_locator(mdates.DayLocator())
    ax.set_xlabel("Date")
    ax.legend(loc='lower right')

    output_path = os.path.join("plot", "binacc.png") \
        if not output_path else output_path
    logging.info(f"Saving to {output_path}")
    plt.savefig(output_path)

    return binacc_fc, binacc_cmp


def compute_sea_ice_extent_error(masks: object,
                                 fc_da: object,
                                 obs_da: object,
                                 grid_area_size: int,
                                 threshold: float) -> object:
    """
    Compute sea ice extent (SIE) error of a forecast, where SIE error is
    defined as the total area covered by grid cells with SIC > (threshold*100)%.

    :param masks: an icenet Masks object
    :param fc_da: the forecasts given as an xarray.DataArray object 
                  with time, xc, yc coordinates
    :param obs_da: the "ground truth" given as an xarray.DataArray object
                   with time, xc, yc coordinates
    :param grid_area_size: the length of the sides of the grid (in km),
                           by default set to 25 (so area of grid is 25*25)
    :param threshold: the SIC threshold of interest (in percentage as a fraction),
                      i.e. threshold is between 0 and 1

    :return: SIE error for forecast as xarray.DataArray object
    """
    grid_area_size = 25 if grid_area_size is None else grid_area_size
    threshold = 0.15 if threshold is None else threshold
    if (threshold < 0) or (threshold > 1):
        raise ValueError("threshold must be a float between 0 and 1")
    
    # obtain mask
    agcm = masks.get_active_cell_da(obs_da)
    
    # binary for observed (i.e. truth)
    binary_obs_da = obs_da > threshold
    binary_obs_weighted_da = binary_obs_da.astype(int).weighted(agcm)

    # binary for forecast
    binary_fc_da = fc_da > threshold
    binary_fc_weighted_da = binary_fc_da.astype(int).weighted(agcm)
    
    # sie error
    forecast_sie_error = (
        binary_fc_weighted_da.sum(['xc', 'yc']) -
        binary_obs_weighted_da.sum(['xc', 'yc'])
    ) * (grid_area_size**2)
    
    return forecast_sie_error


def plot_sea_ice_extent_error(masks: object,
                              fc_da: object,
                              cmp_da: object,
                              obs_da: object,
                              output_path: object,
                              grid_area_size: int = 25,
                              threshold: float = 0.15) -> object:
    """
    Compute and plot sea ice extent (SIE) error of a forecast, where SIE error is
    defined as the total area covered by grid cells with SIC > (threshold*100)%.
    
    :param masks: an icenet Masks object
    :param fc_da: the forecasts given as an xarray.DataArray object 
                  with time, xc, yc coordinates
    :param cmp_da: a comparison forecast / sea ice data given as an 
                   xarray.DataArray object with time, xc, yc coordinates.
                   If None, will ignore plotting a comparison forecast
    :param obs_da: the "ground truth" given as an xarray.DataArray object
                   with time, xc, yc coordinates
    :param output_path: string specifying the path to store the plot
    :param grid_area_size: the length of the sides of the grid (in km),
                           by default set to 25 (so area of grid is 25*25)
    :param threshold: the SIC threshold of interest (in percentage as a fraction),
                      i.e. threshold is between 0 and 1
    
    :return: tuple of (SIE error for forecast (fc_da), SIE error for comparison (cmp_da))
    """
    forecast_sie_error = compute_sea_ice_extent_error(masks=masks,
                                                      fc_da=fc_da,
                                                      obs_da=obs_da,
                                                      grid_area_size=grid_area_size,
                                                      threshold=threshold)
    
    fig, ax = plt.subplots(figsize=(12, 6))
    ax.set_title(f"SIE error comparison ({grid_area_size} km grid resolution) "
                 f"(threshold SIC = {threshold*100}%)")
    ax.plot(forecast_sie_error.time, forecast_sie_error.values, label="IceNet")

    if cmp_da is not None:
        cmp_sie_error = compute_sea_ice_extent_error(masks=masks,
                                                     fc_da=cmp_da,
                                                     obs_da=obs_da,
                                                     grid_area_size=grid_area_size,
                                                     threshold=threshold)
        ax.plot(cmp_sie_error.time, cmp_sie_error.values, label="SEAS")
    else:
        cmp_sie_error = None

    ax.set_ylabel("Sea ice extent error (km)")
    ax.xaxis.set_major_formatter(
        mdates.ConciseDateFormatter(ax.xaxis.get_major_locator()))
    ax.xaxis.set_major_locator(mdates.MonthLocator())
    ax.xaxis.set_minor_locator(mdates.DayLocator())
    ax.set_xlabel("Date")
    ax.legend(loc='lower right')

    output_path = os.path.join("plot", "sie_error.png") \
        if not output_path else output_path
    logging.info(f"Saving to {output_path}")
    plt.savefig(output_path)

    return forecast_sie_error, cmp_sie_error


def compute_metrics(metrics: object,
                    masks: object,
                    fc_da: object,
                    obs_da: object) -> object:
    """
    Computes metrics based on SIC error which are passed in as a list of strings.
    Returns a dictionary where the keys are the metrics,
    and the values are the computed metrics.

    :param metrics: a list of strings
    :param masks: an icenet Masks object
    :param fc_da: an xarray.DataArray object with time, xc, yc coordinates
    :param obs_da: an xarray.DataArray object with time, xc, yc coordinates
    
    :return: dictionary with keys as metric names and values as 
             xarray.DataArray's storing the computed metrics for each forecast
    """
    # check requested metrics have been implemented
    implemented_metrics = ["mae", "mse", "rmse"]
    for metric in metrics:
        if metric not in implemented_metrics:
            raise NotImplementedError(f"{metric} metric has not been implemented. "
                                      f"Please only choose out of {implemented_metrics}.")
    
    # obtain mask
    mask_da = masks.get_active_cell_da(obs_da)
    
    metric_dict = {}
    # compute raw error
    err_da = (fc_da-obs_da)*100
    if "mae" in metrics:
        # compute absolute SIC errors
        abs_err_da = da.fabs(err_da)
        abs_weighted_da = abs_err_da.weighted(mask_da)
    if "mse" in metrics or "rmse" in metrics:
        # compute squared SIC errors
        square_err_da = err_da**2
        square_weighted_da = square_err_da.weighted(mask_da)
        
    for metric in metrics:
        if metric == "mae":
            metric_dict[metric] = abs_weighted_da.mean(dim=['yc', 'xc'])
        elif metric == "mse":
            if "mse" not in metric_dict.keys():
                # might've already been computed if RMSE came first
                metric_dict["mse"] = square_weighted_da.mean(dim=['yc', 'xc'])
        elif metric == "rmse":
            if "mse" not in metric_dict.keys():
                # check if MSE already been computed
                metric_dict["mse"] = square_weighted_da.mean(dim=['yc', 'xc'])
            metric_dict[metric] = da.sqrt(metric_dict["mse"])

    # only return metrics requested (might've computed MSE when computing RMSE)
    return {k: metric_dict[k] for k in metrics}
    
    
def plot_metrics(metrics: object,
                 masks: object,
                 fc_da: object,
                 cmp_da: object,
                 obs_da: object,
                 output_path: object,
                 separate: bool = False) -> object:
    """
    Computes metrics which are passed in as a list of strings,
    and plots them for each forecast.
    Returns a dictionary where the keys are the metrics,
    and the values are the computed metrics.

    :param metrics: a list of strings
    :param masks: an icenet Masks object
    :param fc_da: an xarray.DataArray object with time, xc, yc coordinates
    :param cmp_da: a comparison forecast / sea ice data given as an 
                   xarray.DataArray object with time, xc, yc coordinates.
                   If None, will ignore plotting a comparison forecast
    :param obs_da: an xarray.DataArray object with time, xc, yc coordinates
    :param output_path: string specifying the path to store the plot(s).
                        If separate=True, this should be a directory
    :param separate: bool specifying whether there is a plot created for
                     each metric (True) or not (False), default is False
    
    :return: dictionary with keys as metric names and values as 
             xarray.DataArray's storing the computed metrics for each forecast
    """
    # compute metrics
    fc_metric_dict = compute_metrics(metrics=metrics,
                                     masks=masks,
                                     fc_da=fc_da,
                                     obs_da=obs_da)
    if cmp_da is not None:
        cmp_metric_dict = compute_metrics(metrics=metrics,
                                          masks=masks,
                                          fc_da=cmp_da,
                                          obs_da=obs_da)
    else:
        cmp_metric_dict = None
    
    if separate:
        # produce separate plots for each metric
        for metric in metrics:
            fig, ax = plt.subplots(figsize=(12, 6))
            ax.set_title(f"{metric.upper()} comparison")
            ax.plot(fc_metric_dict[metric].time,
                    fc_metric_dict[metric].values,
                    label="IceNet")
            if cmp_metric_dict is not None:
                ax.plot(cmp_metric_dict[metric].time,
                        cmp_metric_dict[metric].values,
                        label="SEAS")
                
            ax.xaxis.set_major_formatter(
                mdates.ConciseDateFormatter(ax.xaxis.get_major_locator()))
            ax.xaxis.set_major_locator(mdates.MonthLocator())
            ax.xaxis.set_minor_locator(mdates.DayLocator())
            ax.legend(loc='lower right')
            
            outpath = os.path.join("plot", f"{metric}.png") \
                if not output_path else os.path.join(output_path, f"{metric}.png")
            logging.info(f"Saving to {outpath}")
            plt.savefig(outpath)
    else:
        # produce one plot for all metrics
        fig, ax = plt.subplots(figsize=(12, 6))
        ax.set_title("Metric comparison")
        for metric in metrics:
            ax.plot(fc_metric_dict[metric].time,
                    fc_metric_dict[metric].values,
                    label=f"IceNet {metric.upper()}")
            if cmp_metric_dict is not None:
                ax.plot(cmp_metric_dict[metric].time,
                        cmp_metric_dict[metric].values,
                        label=f"SEAS {metric.upper()}",
                        linestyle="dotted")
        
        ax.set_ylabel("SIC (%)")
        ax.xaxis.set_major_formatter(
            mdates.ConciseDateFormatter(ax.xaxis.get_major_locator()))
        ax.xaxis.set_major_locator(mdates.MonthLocator())
        ax.xaxis.set_minor_locator(mdates.DayLocator())
        ax.set_xlabel("Date")
        ax.legend(loc='lower right')
        
        output_path = os.path.join("plot", "metrics.png") \
            if not output_path else output_path
        logging.info(f"Saving to {output_path}")
        plt.savefig(output_path)
    
    return fc_metric_dict, cmp_metric_dict


def compute_metric_as_dataframe(metric: object,
                                masks: object,
                                init_date: object,
                                fc_da: object,
                                obs_da: object,
                                **kwargs) -> pd.DataFrame:
    """
    Computes a metric for each leadtime in a forecast and stores the
    results in a pandas dataframe with columns 'date' (which is the
    initialisation date passed in), 'leadtime' and the metric name(s).
    
    :param metric: string, or list of strings, specifying which metric(s) to compute
    :param masks: an icenet Masks object
    :param init_date: forecast initialisation date which gets
                      added to pandas dataframe (as string, or datetime object)
    :param fc_da: an xarray.DataArray object with time, xc, yc coordinates
    :param obs_da: an xarray.DataArray object with time, xc, yc coordinates
    :param kwargs: any keyword arguments that are required for the computation
                   of the metric, e.g. 'threshold' for SIE error and binary accuracy
                   metrics, or 'grid_area_size' for SIE error metric
    
    :return: computed metric in a pandas dataframe with columns 'date',
             'leadtime' and 'met' for each metric, met, in metric
    """
    if isinstance(metric, str):
        metric = [metric]
    metric_dict = {}
    for met in metric:
        if met in ["mae", "mse", "rmse"]:
            metric_dict[met] = compute_metrics(metrics=[met],
                                               masks=masks,
                                               fc_da=fc_da,
                                               obs_da=obs_da)[met].values
        elif met == "binacc":
            if "threshold" not in kwargs.keys():
                raise KeyError("if met = 'binacc', must pass in argument for threshold")
            metric_dict[met] = compute_binary_accuracy(masks=masks,
                                                       fc_da=fc_da,
                                                       obs_da=obs_da,
                                                       threshold=kwargs["threshold"]).values
        elif met == "sie":
            if "grid_area_size" not in kwargs.keys():
                raise KeyError("if met = 'sie', must pass in argument for grid_area_size")
            if "threshold" not in kwargs.keys():
                raise KeyError("if met = 'sie', must pass in argument for threshold")
            metric_dict[met] = compute_sea_ice_extent_error(masks=masks,
                                                            fc_da=fc_da,
                                                            obs_da=obs_da,
                                                            grid_area_size=kwargs["grid_area_size"],
                                                            threshold=kwargs["threshold"]).values
        else:
            raise NotImplementedError(f"{met} is not implemented")
    
    # create dataframe from metric_dict
    metric_df = pd.DataFrame(metric_dict)
    
    init_date = pd.to_datetime(init_date)
    # compute day of year after first converting year to a non-leap year
    # avoids issue where 2016-03-31 is different to 2015-03-31
    if init_date.strftime("%m-%d") == "02-29":
        # if date is 29th Feb on a leap year, use dayofyear 59
        # (corresponds to 28th Feb in non-leap years)
        dayofyear = 59
    else:
        dayofyear = init_date.replace(year=2001).dayofyear
    month = init_date.month
    # get target dates
    leadtime = list(range(1, len(metric_df.index)+1, 1))
    target_date = pd.Series([init_date + timedelta(days=d) for d in leadtime])
    target_dayofyear = target_date.dt.dayofyear
    # obtain day of year using same method above to avoid any leap-year issues
    target_dayofyear = pd.Series([59 if d.strftime("%m-%d")=="02-29"
                                  else d.replace(year=2001).dayofyear
                                  for d in target_date])
    target_month = target_date.dt.month
    return pd.concat([pd.DataFrame({"date": init_date,
                                    "dayofyear": dayofyear,
                                    "month": month,
                                    "target_date": target_date,
                                    "target_dayofyear": target_dayofyear,
                                    "target_month": target_month,
                                    "leadtime": leadtime}),
                      metric_df],
                      axis=1)


def compute_metrics_leadtime_avg(metric: str,
                                 masks: object,
                                 hemisphere: str,
                                 forecast_file: str,
                                 ecmwf: bool,
                                 data_path: str,
                                 bias_correct: bool = False,
                                 region: tuple = None,
                                 **kwargs) -> object:
    """
    Given forecast file, for each initialisation date in the xarrray.DataArray
    we compute the metric for each leadtime and store the results
    in a pandas dataframe with columns 'date' (specifying the initialisation date),
    'leadtime' and the metric name. This pandas dataframe can then be used
    to average over leadtime to obtain leadtime averaged metrics.
    
    :param metric: string specifying which metric to compute
    :param masks: an icenet Masks object
    :param hemisphere: string, typically either 'north' or 'south'
    :param forecast_file: string specifying a path to a .nc file
    :param ecmwf: bool to indicate whether or not to compare
                  with ECMWF SEAS forecast. If True, will only average
                  over forecasts where the initialisation dates between IceNet
                  and SEAS are the same
    :param data_path: string specifying where to save the metrics dataframe.
                      If None, dataframe is not saved
    :param bias_correct: bool to indicate whether or not to
                         perform a bias correction on SEAS forecast,
                         by default False. Ignored if ecmwf=False
    :param region: region to zoom in to
    :param kwargs: any keyword arguments that are required for the computation
                   of the metric, e.g. 'threshold' for SIE error and binary accuracy
                   metrics, or 'grid_area_size' for SIE error metric
    
    :return: pandas dataframe with columns 'date', 'leadtime' and the metric name.
    """
    # open forecast file
    fc_ds = xr.open_dataset(forecast_file)
    
    if ecmwf:
        # find out what dates cross over with the SEAS5 predictions
        (fc_start_date, fc_end_date) = (fc_ds.time.values.min(), fc_ds.time.values.max())
        dates = get_seas_forecast_init_dates(hemisphere)
        dates = dates[(dates > fc_start_date) & (dates <= fc_end_date)]
        times = [x for x in fc_ds.time.values if x in dates]
        fc_ds = fc_ds.sel(time=times)
    
    logging.info(f"Computing {metric} for {len(fc_ds.time.values)} forecasts")
    # obtain metric for each leadtime at each initialised date in the forecast file
    
    fc_metrics_list = []
    if ecmwf:
        seas_metrics_list = []
    for time in fc_ds.time.values:
        # obtain forecast
        fc = fc_ds.sel(time=slice(time, time))["sic_mean"]
        obs = get_obs_da(hemisphere=hemisphere,
                         start_date=pd.to_datetime(time) + timedelta(days=1),
                         end_date=pd.to_datetime(time) + timedelta(days=int(fc.leadtime.max())))
        fc = filter_ds_by_obs(fc, obs, time)
        
        if ecmwf:
            # obtain SEAS forecast
            seas = get_seas_forecast_da(hemisphere=hemisphere,
                                        date=pd.to_datetime(time),
                                        bias_correct=bias_correct)
            # remove the initialisation date from dataarray
            seas = seas.assign_coords(dict(xc=seas.xc / 1e3, yc=seas.yc / 1e3))
            seas = seas.isel(time=slice(1, None))
        else:
            seas = None
        
        if region is not None:
            seas, fc, obs, masks = process_regions(region,
                                                   [seas, fc, obs, masks])
        
        # compute metrics
        fc_metrics_list.append(compute_metric_as_dataframe(metric=metric,
                                                           masks=masks,
                                                           init_date=time,
                                                           fc_da=fc,
                                                           obs_da=obs,
                                                           **kwargs))
        if seas is not None:
            seas_metrics_list.append(compute_metric_as_dataframe(metric=metric,
                                                                 masks=masks,
                                                                 init_date=time,
                                                                 fc_da=seas,
                                                                 obs_da=obs,
                                                                 **kwargs))
    
    # groupby the leadtime and compute the mean average of the metric
    fc_metric_df = pd.concat(fc_metrics_list)
    fc_metric_df["forecast_name"] = "IceNet"
    if ecmwf:
        seas_metric_df = pd.concat(seas_metrics_list)
        seas_metric_df["forecast_name"] = "SEAS"
        fc_metric_df = pd.concat([fc_metric_df, seas_metric_df])
    
    if data_path is not None:
        logging.info(f"Saving the metric dataframe in {data_path}")
        try:
            fc_metric_df.to_csv(data_path)
        except OSError:
            # don't break if not successful, still return dataframe
            logging.info("Save not successful! Make sure the data_path directory exists")
        
    return fc_metric_df.reset_index(drop=True)


def _parse_day_of_year(dayofyear: int, leapyear: bool = False) -> int:
    """
    Private function which takes in a day of year (integer or float) and returns
    the integer day of year. Useful for ensuring consistency over leap years,
    as dates after March could have different day of years due to leap years.
    For example, 01/03/00 is 60th day in 2000 but 01/03/01 is the 59th day in 2001.
    
    :param dayofyear: integer as int or float type
    :param leapyear: bool to indicate if we want to convert a leapyear dayofyear to
                     non-leapyear
    
    :return: int dayofyear
    """
    if leapyear:
        return (pd.Timestamp("2000-01-01") + timedelta(days=int(dayofyear) - 1)).strftime("%m-%d")
    else:
        return (pd.Timestamp("2001-01-01") + timedelta(days=int(dayofyear) - 1)).strftime("%m-%d")


def _heatmap_ylabels(metrics_df: pd.DataFrame, average_over: str, groupby_col: str) -> object:
    """
    Private function to return the labels for the y-axis in heatmap plots.

    :param metrics_df: pandas dataframe with columns 'date', 'leadtime' and the metric name
    :param average_over: string to specify how to average the metrics.
                         If average_over="all", averages over all possible
                         forecasts and produces line plot.
                         If average_over="month" or "day", averages
                         over the month or day respectively and produces
                         heat map plot.
    :param groupby_col: string to specify how we are grouping the data.
                        If average_over="all", this is typically "dayofyear"
                        or "target_dayofyear".
                        If average_over="month" or "day", this is typically
                        "month" or "target_month".
    
    :return: list of labels for the y-axis
    """
    if average_over == "day":
        # only add labels to the start, end dates
        # and any days that represent the start of months
        days_of_interest = np.array([metrics_df[groupby_col].min(),
                                     1, 32, 60, 91, 121, 152,
                                     182, 213, 244, 274, 305, 335,
                                     metrics_df[groupby_col].max()])
        labels = [_parse_day_of_year(day)
                    if day in days_of_interest else ""
                    for day in sorted(metrics_df[groupby_col].unique())]
    else:
        # find out what months have been plotted and add their names
        month_names = np.array(["Jan", "Feb", "Mar", "Apr", "May", "Jun",
                                "Jul", "Aug", "Sept", "Oct", "Nov", "Dec"])
        labels = [month_names[month-1]
                    for month in sorted(metrics_df[groupby_col].unique())]

    return labels


def standard_deviation_heatmap(metric: str,
                               model_name: str,
                               metrics_df: pd.DataFrame,
                               average_over: str,
                               target_date_avg: bool = False,
                               output_path: str = None,
                               fc_std_metric: pd.DataFrame = None,
                               vmax: float = None,
                               **kwargs) -> object:
    """
    Produces a heatmap plot of the leadtime standard deviation of a metric.

    :param metric: string specifying which metric to compute
    :param model_name: string specifying the name of the model or forecast
    :param metrics_df: pandas dataframe with columns 'date', 'leadtime' and the metric name
    :param average_over: string to specify how to average the metrics.
                         If average_over="all", averages over all possible
                         forecasts and produces line plot.
                         If average_over="month" or "day", averages
                         over the month or day respectively and produces
                         heat map plot.
    :param target_date_avg: bool to indiciate whether or not to average over
                            the target date of forecast (as opposed to averaging
                            over the initialisation date), by default False
    :param output_path: string specifying the path to store the plot, default None
    :param fc_std_metric: pandas dataframe of the standard deviation of the metric over leadtime.
                          If None, this will be computed here
    :param vmax: float specifying maximum to anchor the colourmap. If None,
                 this is inferred from the data
    
    :return: dataframe of the standard deviation of the metric over leadtime
    """
    logging.info(f"Creating standard deviation over leadtime plot for "
                 f"{metric} metric for {model_name} forecasts")
    if average_over == "day":
        groupby_col = "dayofyear"
    else:
        groupby_col = "month"
    if target_date_avg:
        groupby_col = "target_" + groupby_col
    
    if fc_std_metric is None:
        # compute standard deviation of metric
        fc_std_metric = metrics_df.groupby([groupby_col, "leadtime"]).std(numeric_only=True).\
            reset_index().pivot(index=groupby_col, columns="leadtime", values=metric).\
                sort_values(groupby_col, ascending=True)
    n_forecast_days = fc_std_metric.shape[1]
    
    # set ylabel (if average_over == "all"), or legend label (otherwise)
    if metric in ["mae", "mse", "rmse"]:
        ylabel = f"SIC {metric.upper()} (%)"
    elif metric == "binacc":
        ylabel = f"Binary accuracy (%)"
    elif metric == "sie":
        ylabel = f"SIE error (km)"
        
    # plot heatmap of standard deviation for IceNet
    fig, ax = plt.subplots(figsize=(12, 6))
    sns.heatmap(data=fc_std_metric,
                ax=ax,
                vmin=0,
                vmax=vmax,
                cmap="mako_r",
                cbar_kws=dict(label=f"Standard deviation of {ylabel}"))
    
    # y-axis
    labels = _heatmap_ylabels(metrics_df=metrics_df,
                              average_over=average_over,
                              groupby_col=groupby_col)
    ax.set_yticks(np.arange(len(metrics_df[groupby_col].unique()))+0.5)
    ax.set_yticklabels(labels)
    plt.yticks(rotation=0)
    if target_date_avg:
        ax.set_ylabel("Target date of forecast")
    else:
        ax.set_ylabel("Initialisation date of forecast")
    
    # x-axis
    ax.set_xticks(np.arange(30, n_forecast_days, 30))
    ax.set_xticklabels(np.arange(30, n_forecast_days, 30))
    plt.xticks(rotation=0)
    ax.set_xlabel("Lead time (days)")
    
    # add plot title
    (start_date, end_date) = (metrics_df["date"].min().strftime('%d/%m/%Y'),
                              metrics_df["date"].max().strftime('%d/%m/%Y'))
    time_coverage = "\nStandard deviation over a minimum of " + \
        f"{round((metrics_df[groupby_col].value_counts()/n_forecast_days).min())} " + \
        f"forecasts between {start_date} - {end_date}"
    if metric in ["mae", "mse", "rmse"]:
        title = f"{metric.upper()} comparison"
    elif metric == "binacc":
        title = f"Binary accuracy comparison (threshold SIC = {kwargs['threshold'] * 100}%)"
    elif metric == "sie":
        title = f"SIE error comparison ({kwargs['grid_area_size']} km grid resolution, " +\
            f"threshold SIC = {kwargs['threshold'] * 100}%)"
    ax.set_title(title + f" ({model_name})" + time_coverage)

    # save plot
    targ = "target" if target_date_avg and average_over != "all" else "init"
    filename = f"leadtime_averaged_{targ}_{average_over}_{metric}_{model_name}_std.png"
    output_path = os.path.join("plot", filename) \
        if not output_path else output_path
    logging.info(f"Saving to {output_path}")
    plt.savefig(output_path)
    
    return fc_std_metric


def plot_metrics_leadtime_avg(metric: str,
                              masks: object,
                              hemisphere: str,
                              forecast_file: str,
                              ecmwf: bool,
                              output_path: str,
                              average_over: str,
                              plot_std: bool = False,
                              std_mult: float = 1.0,
                              data_path: str = None,
                              target_date_avg: bool = False,
                              bias_correct: bool = False,
                              region: tuple = None,
                              **kwargs) -> object:
    """
    Plots leadtime averaged metrics either using all the forecasts
    in the forecast file, or averaging them over by month or day.
    
    :param metric: string specifying which metric to compute
    :param masks: an icenet Masks object
    :param hemisphere: string, typically either 'north' or 'south'
    :param forecast_file: a path to a .nc file
    :param ecmwf: bool to indicate whether or not to compare
                  with ECMWF SEAS forecast. If True, will only average
                  over forecasts where the initialisation dates between IceNet
                  and SEAS are the same
    :param output_path: string specifying the path to store the plot
    :param average_over: string to specify how to average the metrics.
                         If average_over="all", averages over all possible
                         forecasts and produces line plot.
                         If average_over="month" or "day", averages
                         over the month or day respectively and produces
                         heat map plot.
    :param target_date_avg: bool to indiciate whether or not to average over
                            the target date of forecast (as opposed to averaging
                            over the initialisation date), by default False
    :param data_path: string specifying a CSV file where metrics dataframe
                      could be loaded from. If loading in the dataframe is
                      not possible, it will compute the metrics dataframe
                      and try to save the dataframe
    :param bias_correct: bool to indicate whether or not to
                         perform a bias correction on SEAS forecast,
                         by default False. Ignored if ecmwf=False
    :param region: region to zoom in to
    :param kwargs: any keyword arguments that are required for the computation
                   of the metric, e.g. 'threshold' for SIE error and binary accuracy
                   metrics, or 'grid_area_size' for SIE error metric
    
    :return: pandas dataframe with columns 'date', 'leadtime' and the metric name
    """
    implemented_metrics = ["binacc", "sie", "mae", "mse", "rmse"]
    if metric not in implemented_metrics:
        raise NotImplementedError(f"{metric} metric has not been implemented. "
                                  f"Please only choose out of {implemented_metrics}.")
    if metric == "binacc":
        # add default kwargs
        if "threshold" not in kwargs.keys():
            kwargs["threshold"] = 0.15
    elif metric == "sie":
        # add default kwargs
        if "grid_area_size" not in kwargs.keys():
            kwargs["grid_area_size"] = 25
        if "threshold" not in kwargs.keys():
            kwargs["threshold"] = 0.15
    elif metric in ["mae", "mse", "rmse"]:
        # remove grid_area_size and threshold kwargs if passed
        if "grid_area_size" in kwargs.keys():
            del kwargs["grid_area_size"]
        if "threshold" in kwargs.keys():
            del kwargs["threshold"]
    
    do_compute_metrics = True
    if data_path is not None:
        # loading in precomputed dataframes for the metrics
        logging.info(f"Attempting to read in metrics dataframe from {data_path}")
        try:
            metric_df = pd.read_csv(data_path)
            metric_df["date"] = pd.to_datetime(metric_df["date"])
            do_compute_metrics = False
        except OSError:
            logging.info(f"Couldn't load in dataframe from {data_path}, "
                         f"will compute metric dataframe and try save to {data_path}")

    if do_compute_metrics:
        # computing the dataframes for the metrics
        # will save dataframe in data_path if data_path is not None
        metric_df = compute_metrics_leadtime_avg(metric=metric,
                                                 hemisphere=hemisphere,
                                                 forecast_file=forecast_file,
                                                 ecmwf=ecmwf,
                                                 masks=masks,
                                                 data_path=data_path,
                                                 bias_correct=bias_correct,
                                                 region=region,
                                                 **kwargs)

    fc_metric_df = metric_df[metric_df["forecast_name"] == "IceNet"]
    seas_metric_df = metric_df[metric_df["forecast_name"] == "SEAS"]
    seas_metric_df = seas_metric_df if (len(seas_metric_df) != 0) and ecmwf else None

    logging.info(f"Creating leadtime averaged plot for {metric} metric")
    fig, ax = plt.subplots(figsize=(12, 6))
    (start_date, end_date) = (fc_metric_df["date"].min().strftime('%d/%m/%Y'),
                              fc_metric_df["date"].max().strftime('%d/%m/%Y'))
    
    # set ylabel (if average_over == "all"), or legend label (otherwise)
    if metric in ["mae", "mse", "rmse"]:
        ylabel = f"SIC {metric.upper()} (%)"
    elif metric == "binacc":
        ylabel = f"Binary accuracy (%)"
    elif metric == "sie":
        ylabel = f"SIE error (km)"
    
    if average_over == "all":
        # averaging metric over leadtime for all forecasts
        fc_avg_metric = fc_metric_df.groupby("leadtime").mean(metric).\
            sort_values("leadtime", ascending=True)[metric]
        n_forecast_days = fc_avg_metric.index.max()
        
        # plot leadtime averaged metrics
        ax.plot(fc_avg_metric.index, fc_avg_metric, label="IceNet", color="blue")
        if plot_std:
            # obtaining the standard deviation of the metric
            fc_std_metric = fc_metric_df.groupby("leadtime")[metric].std().\
                sort_index(ascending=True)
            ci = std_mult * fc_std_metric
            ax.fill_between(x=fc_avg_metric.index,
                            y1=fc_avg_metric - ci,
                            y2=fc_avg_metric + ci,
                            color="blue",
                            alpha=0.1)
        if seas_metric_df is not None:
            seas_avg_metric = seas_metric_df.groupby("leadtime").mean(metric).\
                sort_values("leadtime", ascending=True)[metric]
            ax.plot(seas_avg_metric.index, seas_avg_metric, label="SEAS", color="darkorange")
            if plot_std:
                # obtaining the standard deviation of the metric
                seas_std_metric = seas_metric_df.groupby("leadtime")[metric].std().\
                    sort_index(ascending=True)
                ci = std_mult * seas_std_metric
                ax.fill_between(x=seas_avg_metric.index,
                                y1=seas_avg_metric - ci,
                                y2=seas_avg_metric + ci,
                                color="darkorange",
                                alpha=0.1)

        # string to add in plot title
        time_coverage = f"\nAveraged over {len(fc_metric_df['date'].unique())} " + \
            f"forecasts between {start_date} - {end_date}"
        if plot_std:
            time_coverage += f"\n(with +/-{std_mult} standard deviations)"
        ax.set_ylabel(ylabel)
        ax.legend(loc='lower right')
    elif average_over in ["day", "month"]:
        if average_over == "day":
            groupby_col = "dayofyear"
        else:
            groupby_col = "month"
        if target_date_avg:
            groupby_col = "target_" + groupby_col
        
        # compute metric by first grouping the dataframe by groupby_col and leadtime
        fc_avg_metric = fc_metric_df.groupby([groupby_col, "leadtime"]).mean(metric).\
            reset_index().pivot(index=groupby_col, columns="leadtime", values=metric).\
                sort_values(groupby_col, ascending=True)
        n_forecast_days = fc_avg_metric.shape[1]
        
        if seas_metric_df is not None:
            # compute the difference in leadtime average to SEAS forecast
            seas_avg_metric = seas_metric_df.groupby([groupby_col, "leadtime"]).mean(metric).\
                reset_index().pivot(index=groupby_col, columns="leadtime", values=metric).\
                    sort_values(groupby_col, ascending=True)
            heatmap_df_diff = fc_avg_metric - seas_avg_metric
            max = np.nanmax(np.abs(heatmap_df_diff.values))
            
            # plot heatmap of the difference between IceNet and SEAS
            sns.heatmap(data=heatmap_df_diff,
                        ax=ax,
                        vmax=max,
                        vmin=-max,
                        cmap="seismic_r" if metric in ["binacc", "sie"] else "seismic",
                        cbar_kws=dict(label=f"{ylabel} difference between IceNet and SEAS"))

        else:
            # plot heatmap of the leadtime averaged metric when grouped by groupby_col
            sns.heatmap(data=fc_avg_metric, 
                        ax=ax,
                        cmap="inferno" if metric in ["binacc", "sie"] else "inferno_r",
                        cbar_kws=dict(label=ylabel))

        # string to add in plot title
        time_coverage = "\nAveraged over a minimum of " + \
            f"{round((fc_metric_df[groupby_col].value_counts()/n_forecast_days).min())} " + \
            f"forecasts between {start_date} - {end_date}"

        # y-axis
        labels = _heatmap_ylabels(metrics_df=fc_metric_df,
                                  average_over=average_over,
                                  groupby_col=groupby_col)
        ax.set_yticks(np.arange(len(fc_metric_df[groupby_col].unique()))+0.5)
        ax.set_yticklabels(labels)
                
        plt.yticks(rotation=0)
        if target_date_avg:
            ax.set_ylabel("Target date of forecast")
        else:
            ax.set_ylabel("Initialisation date of forecast")
    else:
        raise NotImplementedError(f"averaging over {average_over} not a valid option.")
    
    # add plot title
    if metric in ["mae", "mse", "rmse"]:
        title = f"{metric.upper()} comparison"
    elif metric == "binacc":
        title = f"Binary accuracy comparison (threshold SIC = {kwargs['threshold'] * 100}%)"
    elif metric == "sie":
        title = f"SIE error comparison ({kwargs['grid_area_size']} km grid resolution, " +\
            f"threshold SIC = {kwargs['threshold'] * 100}%)"
    ax.set_title(title + time_coverage)
        
    # x-axis
    ax.set_xticks(np.arange(30, n_forecast_days, 30))
    ax.set_xticklabels(np.arange(30, n_forecast_days, 30))
    plt.xticks(rotation=0)
    ax.set_xlabel("Lead time (days)")
    
    # save plot
    targ = "target" if target_date_avg and average_over != "all" else "init"
    filename = f"leadtime_averaged_{targ}_{average_over}_{metric}" + \
        ("_comp" if seas_metric_df is not None else "") + ".png"
    output_path = os.path.join("plot", filename) \
        if not output_path else output_path
    logging.info(f"Saving to {output_path}")
    plt.savefig(output_path)
    
    if plot_std and average_over in ["day", "month"]:
        # create heapmap for the standard deviation
        if seas_metric_df is not None:
            # compute the standard deviation of the metric for both the forecast and SEAS5
            fc_std_metric = fc_metric_df.groupby([groupby_col, "leadtime"]).std(numeric_only=True).\
                reset_index().pivot(index=groupby_col, columns="leadtime", values=metric).\
                    sort_values(groupby_col, ascending=True)
            seas_std_metric = seas_metric_df.groupby([groupby_col, "leadtime"]).std(numeric_only=True).\
                reset_index().pivot(index=groupby_col, columns="leadtime", values=metric).\
                    sort_values(groupby_col, ascending=True)
            # compute the maximum standard deviation to obtain a common scale
            vmax = np.nanmax([np.nanmax(fc_std_metric.values), np.nanmax(seas_std_metric.values)])
            # create heapmap for the standard deviation
            standard_deviation_heatmap(metric=metric,
                                       model_name="SEAS",
                                       metrics_df=seas_metric_df,
                                       average_over=average_over,
                                       output_path=output_path.replace(".png", "_SEAS_std.png"),
                                       target_date_avg=target_date_avg,
                                       fc_std_metric=seas_std_metric,
                                       vmax=vmax,
                                       **kwargs)
        else:
            # no need to pre-compute the scale for the heatmap, hence
            # pre-computing the standard deviation of the metric is not required
            fc_std_metric = None
            vmax = None
        standard_deviation_heatmap(metric=metric,
                                   model_name="IceNet",
                                   metrics_df=fc_metric_df,
                                   average_over=average_over,
                                   output_path=output_path.replace(".png", "_IceNet_std.png"),
                                   target_date_avg=target_date_avg,
                                   fc_std_metric=fc_std_metric,
                                   vmax=vmax,
                                   **kwargs)

    return fc_metric_df, seas_metric_df


def sic_error_video(fc_da: object,
                    obs_da: object,
                    land_mask: object,
                    output_path: object) -> object:
    """

    :param fc_da:
    :param obs_da:
    :param land_mask:
    :param output_path:
    
    :return: matplotlib animation
    """

    diff = fc_da - obs_da
    fig, maps = plt.subplots(nrows=1,
                             ncols=3,
                             figsize=(16, 6),
                             layout="tight")
    fig.set_dpi(150)

    leadtime = 0
    fc_plot = fc_da.isel(time=leadtime).to_numpy()
    obs_plot = obs_da.isel(time=leadtime).to_numpy()
    diff_plot = diff.isel(time=leadtime).to_numpy()

    upper_bound = np.max([np.abs(np.min(diff_plot)), np.abs(np.max(diff_plot))])
    diff_vmin = -upper_bound
    diff_vmax = upper_bound
    logging.debug("Bounds of differences: {} - {}".format(diff_vmin, diff_vmax))

    sic_cmap = mpl.cm.get_cmap("Blues_r", 20)
    contour_kwargs = dict(
        vmin=0,
        vmax=1,
        cmap=sic_cmap
    )

    diff_cmap = mpl.cm.get_cmap("RdBu_r", 20)
    im1 = maps[0].imshow(fc_plot, **contour_kwargs)
    im2 = maps[1].imshow(obs_plot, **contour_kwargs)
    im3 = maps[2].imshow(diff_plot,
                         vmin=diff_vmin,
                         vmax=diff_vmax,
                         cmap=diff_cmap)

    tic = maps[0].set_title("IceNet "
                            f"{pd.to_datetime(fc_da.isel(time=leadtime).time.values).strftime('%d/%m/%Y')}")
    tio = maps[1].set_title("OSISAF Obs "
                            f"{pd.to_datetime(obs_da.isel(time=leadtime).time.values).strftime('%d/%m/%Y')}")
    maps[2].set_title("Diff")

    p0 = maps[0].get_position().get_points().flatten()
    p1 = maps[1].get_position().get_points().flatten()
    p2 = maps[2].get_position().get_points().flatten()

    ax_cbar = fig.add_axes([p0[0]-0.05, 0.04, p1[2]-p0[0], 0.02])
    plt.colorbar(im1, orientation='horizontal', cax=ax_cbar)

    ax_cbar1 = fig.add_axes([p2[0]+0.05, 0.04, p2[2]-p2[0], 0.02])
    plt.colorbar(im3, orientation='horizontal', cax=ax_cbar1)

    for m_ax in maps[0:3]:
        m_ax.tick_params(
            labelbottom=False,
            labelleft=False,
        )
        m_ax.contourf(land_mask,
                      levels=[.5, 1],
                      colors=[mpl.cm.gray(180)],
                      zorder=3)

    def update(date):
        logging.debug(f"Plotting {date}")

        fc_plot = fc_da.isel(time=date).to_numpy()
        obs_plot = obs_da.isel(time=date).to_numpy()
        diff_plot = diff.isel(time=date).to_numpy()

        tic.set_text("IceNet {}".format(
            pd.to_datetime(fc_da.isel(time=date).time.values).strftime("%d/%m/%Y")))
        tio.set_text("OSISAF Obs {}".format(
            pd.to_datetime(obs_da.isel(time=date).time.values).strftime("%d/%m/%Y")))

        im1.set_data(fc_plot)
        im2.set_data(obs_plot)
        im3.set_data(diff_plot)

        return tic, tio, im1, im2, im3

    animation = FuncAnimation(fig,
                              update,
                              range(0, len(fc_da.time)),
                              interval=100)

    plt.close()

    output_path = os.path.join("plot", "sic_error.mp4") \
        if not output_path else output_path
    logging.info(f"Saving to {output_path}")
    animation.save(output_path,
                   fps=10,
                   extra_args=['-vcodec', 'libx264'])
    return animation


def sic_error_local_header_data(da: xr.DataArray):
    n_probe = len(da.probe)
    return {
        "probe array index": {
            i_probe: (
                f"{da.xi.values[i_probe]},"
                f"{da.yi.values[i_probe]}"
            )
            for i_probe in range(n_probe)
        },
        "probe location (EASE)": {
            i_probe: (
                f"{da.xc.values[i_probe]},"
                f"{da.yc.values[i_probe]}"
            )
            for i_probe in range(n_probe)
        },
        "probe location (lat, lon)": {
            i_probe: (
                f"{da.lat.values[i_probe]},"
                f"{da.lon.values[i_probe]}"
            )
            for i_probe in range(n_probe)
        },
        "obs_kind": {
            0: "forecast",
            1: "observation",
            2: "forecast error ('0' - '1')",
            # optionally, include SEAS comparison
        },
    }


def sic_error_local_write_fig(combined_da: xr.DataArray,
                              output_prefix: str):
    """A helper function for `sic_error_local_plots`: plot error and
    forecast/observation data.

    :param combined_da: A DataArray with dims ('time', 'probe', 'obs_kind')

    :param output_prefix: A string from which to produce the output
    path (the probe index and file extension will be appended).
    """

    OBS_KIND_FC = 0
    OBS_KIND_OBS = 1
    OBS_KIND_ERR = 2

    plot_series = combined_da.to_dataframe(name="SIC")["SIC"]

    outfile = output_prefix + ".fc.pdf"

    all_figs = []
    with PdfPages(outfile) as output_pdf:
        n_probe = len(combined_da.probe)
        for i_probe in range(n_probe):

            #### Forecast/observed plots ####

            lat = combined_da.probe.lat.values[i_probe]
            lon = combined_da.probe.lon.values[i_probe]

            lat_h = "N" if lat >= 0.0 else "S"
            lat = abs(lat)

            lon_h = "E" if lon >= 0.0 else "W"
            lon = abs(lon)

            fig, ax = plt.subplots()
            all_figs.append(fig)

            ax.set_title(
                f"Sea ice concentration at location {i_probe + 1}\n"
                f"{lat:.3f}° {lat_h}, {lon:.3f}° {lon_h}"
            )

            ax.set_xlabel("Date")
            ax.set_ylabel("SIC (%)")

            # dims: (obs_kind, time, probe)
            ax.plot(plot_series.loc[OBS_KIND_FC, :, i_probe], label="IceNet")
            ax.plot(plot_series.loc[OBS_KIND_OBS, :, i_probe], label="Observed")
            ax.legend()

            plt.setp(ax.get_xticklabels(), rotation=45, ha='right')
            output_pdf.savefig(fig, bbox_inches='tight')

            #### Error plot ####
            fig2, ax2 = plt.subplots()
            all_figs.append(fig2)

            ax2.set_title(
                f"Sea ice concentration error at location {i_probe + 1}\n"
                f"{lat:.3f}° {lat_h}, {lon:.3f}° {lon_h}"
            )

            ax2.set_xlabel("Date")
            ax2.set_ylabel("SIC error (%)")

            ax2.axhline(color='k', lw=0.5, ls='--')
            ax2.plot(plot_series.loc[OBS_KIND_ERR, :, i_probe], color='C2')

            plt.setp(ax2.get_xticklabels(), rotation=45, ha='right')
            output_pdf.savefig(fig2, bbox_inches='tight')

    return all_figs


def sic_error_local_plots(fc_da: object,
                          obs_da: object,
                          output_path: object,
                          as_command: bool = False):

    """
    :param fc_da: a DataArray with dims ('time', 'probe')
    :param obs_da: a DataArray with dims ('time', 'probe')
    """

    # convert SIC to percentages (ranging from 0% to 100%) rather than a fraction
    fc_da = fc_da*100
    obs_da = obs_da*100
    err_da = (fc_da-obs_da)
    combined_da = xr.concat(
        [fc_da, obs_da, err_da],
        dim="obs_kind", coords="minimal"
    )

    # convert to a dataframe for csv output
    df = (
        combined_da
        .to_dataframe(name="SIC")
        # drop unneeded coords (lat, lon, xc, yc)
        .loc[:, "SIC"]
        # Convert mult-indices
        .unstack(2).unstack(0)
    )

    if output_path is None:
        output_path = "sic_error_local.csv"

    header_info = sic_error_local_header_data(combined_da)

    header = "# icenet_plot_sic_error_local\n"
    header += f"# Part of Icenet, version {icenet_version}\n"
    header += "#\n"

    if as_command:
        header += "# Command output from \n"
        cmd = ' '.join([a.__repr__() for a in sys.argv])
        header += f"#   {cmd}\n"
        header += "#\n"

    header += "# Key\n"
    for header_kind, header_data in header_info.items():
        header += f"# {header_kind}\n"
        for k, v in header_data.items():
            header += f"#   {k}: {v}\n"

    with open(output_path, "w") as outfile:
        outfile.write(header)
        df.to_csv(outfile)

    figs = sic_error_local_write_fig(combined_da, "sic_error_local")

    return figs


class ForecastPlotArgParser(argparse.ArgumentParser):

    """An ArgumentParser specialised to support forecast plot arguments

    Additional argument enabled by allow_ecmwf() etc.

    The 'allow_*' methods return self to permit method chaining.

    :param forecast_date: allows this positional argument to be disabled
    """

    def __init__(self, *args,
                 forecast_date: bool = True,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self.add_argument("hemisphere", choices=("north", "south"))
        self.add_argument("forecast_file", type=str)
        if forecast_date:
            self.add_argument("forecast_date", type=date_arg)

        self.add_argument("-o", "--output-path", type=str, default=None)
        self.add_argument("-v", "--verbose", action="store_true", default=False)
        self.add_argument("-r", "--region", default=None, type=region_arg,
                          help="Region specified x1, y1, x2, y2")

    def allow_ecmwf(self):
        self.add_argument("-b", "--bias-correct",
                          help="Bias correct SEAS forecast array",
                          action="store_true",
                          default=False)
        self.add_argument("-e", "--ecmwf", action="store_true", default=False)
        return self

    def allow_threshold(self):
        self.add_argument("-t",
                          "--threshold",
                          help="The SIC threshold of interest",
                          type=float,
                          default=0.15)
        return self

    def allow_sie(self):
        self.add_argument("-ga",
                          "--grid-area",
                          help="The length of the sides of the grid used (in km)",
                          type=int,
                          default=25)
        return self

    def allow_metrics(self):
        self.add_argument("-m", 
                          "--metrics",
                          help="Which metrics to compute and plot",
                          type=str,
                          default="mae,mse,rmse")
        self.add_argument("-s",
                          "--separate",
                          help="Whether or not to produce separate plots for each metric",
                          action="store_true",
                          default=False)
        return self

    def allow_probes(self):
        self.add_argument(
            "-p", "--probe", action="append", dest="probes",
            type=location_arg, metavar="LOCATION",
            help="Sample at LOCATION",
        )
        return self

    def parse_args(self, *args, **kwargs):
        args = super().parse_args(*args, **kwargs)

        logging.basicConfig(level=logging.DEBUG if args.verbose else logging.INFO)
        logging.getLogger("matplotlib").setLevel(logging.WARNING)

        return args

##
# CLI endpoints
#


def binary_accuracy():
    """
    Produces plot of the binary classification accuracy of forecasts.
    """
    ap = (
        ForecastPlotArgParser()
        .allow_ecmwf()
        .allow_threshold()
    )
    args = ap.parse_args()

    masks = Masks(north=args.hemisphere == "north",
                  south=args.hemisphere == "south")

    fc = get_forecast_ds(args.forecast_file,
                         args.forecast_date)
    obs = get_obs_da(args.hemisphere,
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=1),
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=int(fc.leadtime.max())))
    fc = filter_ds_by_obs(fc, obs, args.forecast_date)

    if args.ecmwf:
        seas = get_seas_forecast_da(
            args.hemisphere,
            args.forecast_date,
            bias_correct=args.bias_correct) \
            if args.ecmwf else None

        if seas is not None:
            seas = seas.assign_coords(dict(xc=seas.xc / 1e3, yc=seas.yc / 1e3))
            seas = seas.isel(time=slice(1, None))
    else:
        seas = None

    if args.region:
        seas, fc, obs, masks = process_regions(args.region,
                                               [seas, fc, obs, masks])
    
    plot_binary_accuracy(masks=masks,
                         fc_da=fc,
                         cmp_da=seas,
                         obs_da=obs,
                         output_path=args.output_path,
                         threshold=args.threshold)


def sie_error():
    """
    Produces plot of the sea ice extent (SIE) error of forecasts.
    """
    ap = (
        ForecastPlotArgParser()
        .allow_ecmwf()
        .allow_threshold()
        .allow_sie()
    )
    args = ap.parse_args()

    masks = Masks(north=args.hemisphere == "north",
                  south=args.hemisphere == "south")

    fc = get_forecast_ds(args.forecast_file,
                         args.forecast_date)
    obs = get_obs_da(args.hemisphere,
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=1),
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=int(fc.leadtime.max())))
    fc = filter_ds_by_obs(fc, obs, args.forecast_date)

    if args.ecmwf:
        seas = get_seas_forecast_da(
            args.hemisphere,
            args.forecast_date,
            bias_correct=args.bias_correct) \
            if args.ecmwf else None

        if seas is not None:
            seas = seas.assign_coords(dict(xc=seas.xc / 1e3, yc=seas.yc / 1e3))
            seas = seas.isel(time=slice(1, None))
    else:
        seas = None

    if args.region:
        seas, fc, obs, masks = process_regions(args.region,
                                               [seas, fc, obs, masks])

    plot_sea_ice_extent_error(masks=masks,
                              fc_da=fc,
                              cmp_da=seas,
                              obs_da=obs,
                              output_path=args.output_path,
                              grid_area_size=args.grid_area,
                              threshold=args.threshold)


def plot_forecast():
    """CLI entry point for icenet_plot_forecast

    :return:
    """
    ap = ForecastPlotArgParser()
    ap.add_argument("-l", "--leadtimes",
                    help="Leadtimes to output, multiple as CSV, range as n..n",
                    type=lambda s: [int(i) for i in
                                    list(s.split(",") if "," in s else
                                         range(int(s.split("..")[0]),
                                               int(s.split("..")[1]) + 1) if ".." in s else
                                         [s])])
    ap.add_argument("-c", "--no-coastlines",
                    help="Turn off cartopy integration",
                    action="store_true", default=False)
    ap.add_argument("-f", "--format",
                    help="Format to output in",
                    choices=("mp4", "png", "svg", "tiff"),
                    default="png")
    ap.add_argument("-n", "--cmap-name",
                    help="Color map name if not wanting to use default",
                    default=None)
    ap.add_argument("-s", "--stddev",
                    help="Plot the standard deviation from the ensemble",
                    action="store_true",
                    default=False)
    args = ap.parse_args()

    fc = get_forecast_ds(args.forecast_file,
                         args.forecast_date,
                         stddev=args.stddev)
    fc = fc.transpose(..., "yc", "xc")

    output_path = "." if args.output_path is None else args.output_path

    if not os.path.isdir(output_path):
        logging.warning("No directory at: {}".format(output_path))
        os.makedirs(output_path)
    elif os.path.isfile(output_path):
        raise RuntimeError("{} should be a directory and not existent...".
                           format(output_path))

    forecast_name = "{}.{}".format(
        os.path.splitext(os.path.basename(args.forecast_file))[0],
        args.forecast_date)

    cmap_name = "BuPu_r" if args.stddev else "Blues_r"
    if args.cmap_name is not None:
        cmap_name = args.cmap_name

    logging.info("Using cmap {}".format(cmap_name))
    cmap = cm.get_cmap(cmap_name)
    cmap.set_bad("dimgrey")

    if args.region is not None:
        fc = process_regions(args.region, [fc])[0]

    vmax = 1.

    if args.stddev:
        vmax = float(fc.max())
        logging.info("Calculated vmax to be: {}".format(vmax))

    leadtimes = args.leadtimes \
        if args.leadtimes is not None \
        else list(range(1, int(max(fc.leadtime.values)) + 1))

    if args.format == "mp4":
        pred_da = fc.isel(time=0).sel(leadtime=leadtimes)

        if "forecast_date" not in pred_da:
            forecast_dates = [
                pd.Timestamp(args.forecast_date) + dt.timedelta(lt)
                for lt in args.leadtimes]
            pred_da = pred_da.assign_coords(
                forecast_date=("leadtime", forecast_dates))

        pred_da = pred_da.drop("time").drop("leadtime").\
            rename(leadtime="time", forecast_date="time").set_index(time="time")

        anim_args = dict(
            figsize=5
        )
        if not args.no_coastlines:
            logging.warning("Coastlines will not work with the current "
                            "implementation of xarray_to_video")

        output_filename = os.path.join(output_path, "{}.{}.{}{}".format(
            forecast_name,
            args.forecast_date.strftime("%Y%m%d"),
            "" if not args.stddev else "stddev.",
            args.format
        ))
        xarray_to_video(pred_da, fps=1, cmap=cmap,
                        imshow_kwargs=dict(vmin=0., vmax=vmax)
                        if not args.stddev else None,
                        video_path=output_filename,
                        **anim_args)
    else:
        for leadtime in leadtimes:
            pred_da = fc.sel(leadtime=leadtime).isel(time=0)
            bound_args = dict(north=args.hemisphere == "north",
                              south=args.hemisphere == "south")

            if args.region is not None:
                bound_args.update(x1=args.region[0],
                                  x2=args.region[2],
                                  y1=args.region[1],
                                  y2=args.region[3])

            ax = get_plot_axes(**bound_args,
                               do_coastlines=not args.no_coastlines)

            bound_args.update(cmap=cmap)

            im = show_img(ax, pred_da, **bound_args, vmax=vmax,
                          do_coastlines=not args.no_coastlines)

            plt.colorbar(im, ax=ax)
            plot_date = args.forecast_date + dt.timedelta(leadtime)
            ax.set_title("{:04d}/{:02d}/{:02d}".format(plot_date.year,
                                                       plot_date.month,
                                                       plot_date.day))
            output_filename = os.path.join(output_path, "{}.{}.{}{}".format(
                forecast_name,
                (args.forecast_date + dt.timedelta(
                    days=leadtime)).strftime("%Y%m%d"),
                "" if not args.stddev else "stddev.",
                args.format
            ))

            logging.info("Saving to {}".format(output_filename))
            plt.savefig(output_filename)
            plt.clf()


def parse_metrics_arg(argument: str) -> object:
    """
    Splits a string into a list by separating on commas.
    Will remove any whitespace and removes duplicates.
    Used to parsing metrics argument in metric_plots.
    
    :param argument: string
    
    :return: list of metrics to compute
    """
    return list(set([s.replace(" ", "") for s in argument.split(",")]))


def metric_plots():
    """
    Produces plot of requested metrics for forecasts.
    """
    ap = (
        ForecastPlotArgParser()
        .allow_ecmwf()
        .allow_metrics()
    )
    args = ap.parse_args()

    masks = Masks(north=args.hemisphere == "north",
                  south=args.hemisphere == "south")

    fc = get_forecast_ds(args.forecast_file,
                         args.forecast_date)
    obs = get_obs_da(args.hemisphere,
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=1),
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=int(fc.leadtime.max())))
    fc = filter_ds_by_obs(fc, obs, args.forecast_date)

    metrics = parse_metrics_arg(args.metrics)

    if args.ecmwf:
        seas = get_seas_forecast_da(
            args.hemisphere,
            args.forecast_date,
            bias_correct=args.bias_correct) \
            if args.ecmwf else None

        if seas is not None:
            seas = seas.assign_coords(dict(xc=seas.xc / 1e3, yc=seas.yc / 1e3))
            seas = seas.isel(time=slice(1, None))
    else:
        seas = None

    if args.region:
        seas, fc, obs, masks = process_regions(args.region,
                                               [seas, fc, obs, masks])

    plot_metrics(metrics=metrics,
                 masks=masks,
                 fc_da=fc,
                 cmp_da=seas,
                 obs_da=obs,
                 output_path=args.output_path,
                 separate=args.separate)


def leadtime_avg_plots():
    """
    Produces plot of leadtime averaged metrics for forecasts.
    """
    ap = (
        ForecastPlotArgParser(forecast_date=False)
        .allow_ecmwf()
        .allow_threshold()
        .allow_sie()
    )
    ap.add_argument("-m",
                    "--metric",
                    help="Which metric to compute and plot",
                    type=str)
    ap.add_argument("-dp",
                    "--data_path",
                    help="Where to find (or store) metrics dataframe",
                    type=str,
                    default=None)
    ap.add_argument("-ao",
                    "--average_over",
                    help="How to average the forecast metrics",
                    type=str,
                    choices=["all", "month", "day"])
    ap.add_argument("-s",
                    "--std",
                    help="If flagged, standard deviation is plotted",
                    action="store_true",
                    default=False)
    ap.add_argument("-sm",
                    "--std_mult",
                    help="What multiple of the standard deviation to plot",
                    type=float,
                    default=1.0)
    ap.add_argument("-td",
                    "--target_date_average",
                    help="Averages metric over target date instead of init date",
                    action="store_true",
                    default=False)
    args = ap.parse_args()
    masks = Masks(north=args.hemisphere == "north",
                  south=args.hemisphere == "south")
    
    plot_metrics_leadtime_avg(metric=args.metric,
                              masks=masks,
                              hemisphere=args.hemisphere,
                              forecast_file=args.forecast_file,
                              ecmwf=args.ecmwf,
                              output_path=args.output_path,
                              average_over=args.average_over,
                              plot_std=args.std,
                              std_mult=args.std_mult,
                              data_path=args.data_path,
                              target_date_avg=args.target_date_average,
                              bias_correct=args.bias_correct,
                              region=args.region,
                              threshold=args.threshold,
                              grid_area_size=args.grid_area)


def sic_error():
    """
    Produces video visualisation of SIC of forecast and ground truth.
    """
    ap = ForecastPlotArgParser()
    args = ap.parse_args()

    masks = Masks(north=args.hemisphere == "north",
                  south=args.hemisphere == "south")

    fc = get_forecast_ds(args.forecast_file,
                         args.forecast_date)
    obs = get_obs_da(args.hemisphere,
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=1),
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=int(fc.leadtime.max())))
    fc = filter_ds_by_obs(fc, obs, args.forecast_date)

    if args.region:
        fc, obs, masks = process_regions(args.region, [fc, obs, masks])

    sic_error_video(fc_da=fc,
                    obs_da=obs,
                    land_mask=masks.get_land_mask(),
                    output_path=args.output_path)


def sic_error_local():
    """
    Entry point for the icenet_plot_sic_error_local command
    """

    ap = (
        ForecastPlotArgParser()
        .allow_probes()
    )
    args = ap.parse_args()

    fc = get_forecast_ds(args.forecast_file,
                         args.forecast_date)
    obs = get_obs_da(args.hemisphere,
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=1),
                     pd.to_datetime(args.forecast_date) +
                     timedelta(days=int(fc.leadtime.max())))
    fc = filter_ds_by_obs(fc, obs, args.forecast_date)

    fc, obs = process_probes(args.probes, [fc, obs])

    sic_error_local_plots(fc,
                          obs,
                          args.output_path,
                          as_command=True)



"""Top-level package for IceNet."""

__author__ = "British Antarctic Survey / Alan Turing Institute"
__copyright__ = "British Antarctic Survey"
__email__ = "jambyr@bas.ac.uk"
__license__ = "MIT"
__version__ = "0.2.6"



import datetime as dt
import glob
import logging
import os
import re

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import pandas as pd
import xarray as xr


from ibicus.debias import LinearScaling


def broadcast_forecast(start_date: object,
                       end_date: object,
                       datafiles: object = None,
                       dataset: object = None,
                       target: object = None) -> object:
    """

    :param start_date:
    :param end_date:
    :param datafiles:
    :param dataset:
    :param target:
    :return:
    """

    assert (datafiles is None) ^ (dataset is None), \
        "Only one of datafiles and dataset can be set"

    if datafiles:
        logging.info("Using {} to generate forecast through {} to {}".
                     format(", ".join(datafiles), start_date, end_date))
        dataset = xr.open_mfdataset(datafiles, engine="netcdf4")

    dates = pd.date_range(start_date, end_date)
    i = 0

    logging.debug("Dataset summary: \n{}".format(dataset))

    if len(dataset.time.values) > 1:
        while dataset.time.values[i + 1] < dates[0]:
            i += 1

    logging.info("Starting index will be {} for {} - {}".
                 format(i, dates[0], dates[-1]))
    dt_arr = []

    for d in dates:
        logging.debug("Looking for date {}".format(d))
        arr = None

        while arr is None:
            if d >= dataset.time.values[i]:
                d_lead = (d - dataset.time.values[i]).days

                if i + 1 < len(dataset.time.values):
                    if pd.to_datetime(dataset.time.values[i]) + \
                            dt.timedelta(days=d_lead) >= \
                            pd.to_datetime(dataset.time.values[i + 1]) + \
                            dt.timedelta(days=1):
                        i += 1
                        continue

                logging.debug("Selecting date {} and lead {}".
                              format(pd.to_datetime(
                                     dataset.time.values[i]).strftime("%D"),
                                     d_lead))

                arr = dataset.sel(time=dataset.time.values[i],
                                  leadtime=d_lead).\
                    copy().\
                    drop("time").\
                    assign_coords(dict(time=d)).\
                    drop("leadtime")
            else:
                i += 1

        dt_arr.append(arr)

    target_ds = xr.concat(dt_arr, dim="time")

    if target:
        logging.info("Saving dataset to {}".format(target))
        target_ds.to_netcdf(target)
    return target_ds


def get_seas_forecast_init_dates(hemisphere: str,
                                 source_path: object = os.path.join(".", "data", "mars.seas")) -> object:
    """
    Obtains list of dates for which we have SEAS forecasts we have.

    :param hemisphere: string, typically either 'north' or 'south'
    :param source_path: path where north and south SEAS forecasts are stored
    
    :return: list of dates
    """
    # list the files in the path where SEAS forecasts are stored
    filenames = os.listdir(os.path.join(source_path, 
                                        hemisphere,
                                        "siconca"))
    # obtain the dates from files with YYYYMMDD.nc format
    return pd.to_datetime([x.split('.')[0]
                           for x in filenames
                           if re.search(r'^\d{8}\.nc$', x)])


def get_seas_forecast_da(hemisphere: str,
                         date: str,
                         bias_correct: bool = True,
                         source_path: object = os.path.join(".", "data", "mars.seas"),
                         ) -> tuple:
    """
    Atmospheric model Ensemble 15-day forecast (Set III - ENS)

Coordinates:
  * time                          (time) datetime64[ns] 2022-04-01 ... 2022-0...
  * yc                            (yc) float64 5.388e+06 ... -5.388e+06
  * xc                            (xc) float64 -5.388e+06 ... 5.388e+06

    :param hemisphere: string, typically either 'north' or 'south'
    :param date:
    :param bias_correct:
    :param source_path:
    """

    seas_file = os.path.join(
        source_path,
        hemisphere,
        "siconca",
        "{}.nc".format(date.replace(day=1).strftime("%Y%m%d")))

    if os.path.exists(seas_file):
        seas_da = xr.open_dataset(seas_file).siconc
    else:
        logging.warning("No SEAS data available at {}".format(seas_file))
        return None

    if bias_correct:
        # Let's have some maximum, though it's quite high
        (start_date, end_date) = (
            date - dt.timedelta(days=10 * 365),
            date + dt.timedelta(days=10 * 365)
        )
        obs_da = get_obs_da(hemisphere, start_date, end_date)
        seas_hist_files = dict(sorted({os.path.abspath(el):
                                       dt.datetime.strptime(
                                       os.path.basename(el)[0:8], "%Y%m%d")
                                      for el in
                                      glob.glob(os.path.join(source_path,
                                                             hemisphere,
                                                             "siconca",
                                                             "*.nc"))
                                      if re.search(r'^\d{8}\.nc$',
                                                   os.path.basename(el))
                                      and el != seas_file}.items()))

        def strip_overlapping_time(ds):
            data_file = os.path.abspath(ds.encoding["source"])

            try:
                idx = list(seas_hist_files.keys()).index(data_file)
            except ValueError:
                logging.exception("\n{} not in \n\n{}".format(data_file,
                                                              seas_hist_files))
                return None

            if idx < len(seas_hist_files) - 1:
                max_date = seas_hist_files[
                               list(seas_hist_files.keys())[idx + 1]] \
                           - dt.timedelta(days=1)
                logging.debug("Stripping {} to {}".format(data_file, max_date))
                return ds.sel(time=slice(None, max_date))
            else:
                logging.debug("Not stripping {}".format(data_file))
                return ds

        hist_da = xr.open_mfdataset(seas_hist_files,
                                    preprocess=strip_overlapping_time).siconc
        debiaser = LinearScaling(delta_type="additive",
                                 variable="siconc",
                                 reasonable_physical_range=[0., 1.])

        logging.info("Debiaser input ranges: obs {:.2f} - {:.2f}, "
                     "hist {:.2f} - {:.2f}, fut {:.2f} - {:.2f}".
                     format(float(obs_da.min()), float(obs_da.max()),
                            float(hist_da.min()), float(hist_da.max()),
                            float(seas_da.min()), float(seas_da.max())))

        seas_array = debiaser.apply(obs_da.values,
                                    hist_da.values,
                                    seas_da.values)
        seas_da.values = seas_array
        logging.info("Debiaser output range: {:.2f} - {:.2f}".
                     format(float(seas_da.min()), float(seas_da.max())))

    logging.info("Returning SEAS data from {} from {}".format(seas_file, date))

    # This isn't great looking, but we know we're not dealing with huge
    # indexes in here
    date_location = list(seas_da.time.values).index(pd.Timestamp(date))
    if date_location > 0:
        logging.warning("SEAS forecast started {} day before the requested "
                        "date {}, make sure you account for this!".
                        format(date_location, date))

    seas_da = seas_da.sel(time=slice(date, None))
    logging.debug("SEAS data range: {} - {}, {} dates".format(
        pd.to_datetime(min(seas_da.time.values)).strftime("%Y-%m-%d"),
        pd.to_datetime(max(seas_da.time.values)).strftime("%Y-%m-%d"),
        len(seas_da.time)
    ))
    
    return seas_da


def get_forecast_ds(forecast_file: object,
                    forecast_date: str,
                    stddev: bool = False
                    ) -> object:
    """

    :param forecast_file: a path to a .nc file
    :param forecast_date: initialisation date of the forecast
    :param stddev:
    :returns tuple(fc_ds, obs_ds, land_mask):
    """
    forecast_date = pd.to_datetime(forecast_date)

    forecast_ds = xr.open_dataset(forecast_file, decode_coords="all")
    get_key = "sic_mean" if not stddev else "sic_stddev"

    forecast_ds = getattr(
        forecast_ds.sel(time=slice(forecast_date, forecast_date)),
        get_key)

    return forecast_ds


def filter_ds_by_obs(ds: object,
                     obs_da: object,
                     forecast_date: str) -> object:
    """

    :param ds:
    :param obs_da:
    :param forecast_date: initialisation date of the forecast
    :return:
    """
    forecast_date = pd.to_datetime(forecast_date)
    (start_date, end_date) = (
            forecast_date + dt.timedelta(days=int(ds.leadtime.min())),
            forecast_date + dt.timedelta(days=int(ds.leadtime.max()))
    )

    if len(obs_da.time) < len(ds.leadtime):
        if len(obs_da.time) < 1:
            raise RuntimeError("No observational data available between {} "
                               "and {}".format(start_date.strftime("%D"),
                                               end_date.strftime("%D")))

        logging.warning("Observational data not available for full range of "
                        "forecast lead times: {}-{} vs {}-{}".format(
                         obs_da.time.to_series()[0].strftime("%D"),
                         obs_da.time.to_series()[-1].strftime("%D"),
                         start_date.strftime("%D"),
                         end_date.strftime("%D")))
        (start_date, end_date) = (
            obs_da.time.to_series()[0],
            obs_da.time.to_series()[-1]
        )

    # We broadcast to get a nicely compatible dataset for plotting
    return broadcast_forecast(start_date=start_date,
                              end_date=end_date,
                              dataset=ds)


def get_obs_da(hemisphere: str,
               start_date: str,
               end_date: str,
               obs_source: object =
               os.path.join(".", "data", "osisaf"),
               ) -> object:
    """

    :param hemisphere: string, typically either 'north' or 'south'
    :param start_date:
    :param end_date:
    :param obs_source:
    :return:
    """
    obs_years = pd.Series(pd.date_range(start_date, end_date)).dt.year.unique()
    obs_dfs = [el for yr in obs_years for el in
               glob.glob(os.path.join(obs_source,
                                      hemisphere,
                                      "siconca", "{}.nc".format(yr)))]

    if len(obs_dfs) < len(obs_years):
        logging.warning("Cannot find all obs source files for {} - {} in {}".
                        format(start_date, end_date, obs_source))

    logging.info("Got files: {}".format(obs_dfs))
    obs_ds = xr.open_mfdataset(obs_dfs)
    obs_ds = obs_ds.sel(time=slice(start_date, end_date))

    return obs_ds.ice_conc


def calculate_extents(x1: int,
                      x2: int,
                      y1: int,
                      y2: int):
    """

    :param x1:
    :param x2:
    :param y1:
    :param y2:
    :return:
    """
    data_extent_base = 5387500

    extents = [
        -data_extent_base + (x1 * 25000),
        data_extent_base - ((432 - x2) * 25000),
        -data_extent_base + (y1 * 25000),
        data_extent_base - ((432 - y2) * 25000),
    ]

    logging.debug("Data extents: {}".format(extents))
    return extents


def get_plot_axes(x1: int = 0,
                  x2: int = 432,
                  y1: int = 0,
                  y2: int = 432,
                  do_coastlines: bool = True,
                  north: bool = True,
                  south: bool = False):
    """

    :param x1:
    :param x2:
    :param y1:
    :param y2:
    :param do_coastlines:
    :param north:
    :param south:
    :return:
    """
    assert north ^ south, "One hemisphere only must be selected"

    fig = plt.figure(figsize=(10, 8), dpi=150, layout='tight')

    if do_coastlines:
        pole = 1 if north else -1
        proj = ccrs.LambertAzimuthalEqualArea(0, pole * 90)
        ax = fig.add_subplot(1, 1, 1, projection=proj)
        extents = calculate_extents(x1, x2, y1, y2)
        ax.set_extent(extents, crs=proj)
    else:
        ax = fig.add_subplot(1, 1, 1)

    return ax


def show_img(ax,
             arr,
             x1: int = 0,
             x2: int = 432,
             y1: int = 0,
             y2: int = 432,
             cmap: object = None,
             do_coastlines: bool = True,
             vmin: float = 0.,
             vmax: float = 1.,
             north: bool = True,
             south: bool = False):
    """

    :param ax:
    :param arr:
    :param x1:
    :param x2:
    :param y1:
    :param y2:
    :param cmap:
    :param do_coastlines:
    :param vmin:
    :param vmax:
    :param north:
    :param south:
    :return:
    """

    assert north ^ south, "One hemisphere only must be selected"

    if do_coastlines:
        pole = 1 if north else -1
        data_crs = ccrs.LambertAzimuthalEqualArea(0, pole * 90)
        extents = calculate_extents(x1, x2, y1, y2)
        im = ax.imshow(arr,
                       vmin=vmin,
                       vmax=vmax,
                       cmap=cmap,
                       transform=data_crs,
                       extent=extents)
        ax.coastlines()
    else:
        im = ax.imshow(arr, cmap=cmap, vmin=vmin, vmax=vmax)

    return im


def process_probes(probes, data) -> tuple:
    """
    :param probes: A sequence of locations (pairs)
    :param data: A sequence of xr.DataArray
    """

    # index into each element of data with a xr.DataArray, for pointwise
    # selection.  Construct the indexing DataArray as follows:

    probes_da = xr.DataArray(probes, dims=('probe', 'coord'))
    xcs, ycs = probes_da.sel(coord=0), probes_da.sel(coord=1)

    for idx, arr in enumerate(data):
        arr = arr.assign_coords({
            "xi": ("xc", np.arange(len(arr.xc))),
            "yi": ("yc", np.arange(len(arr.yc))),
        })
        if arr is not None:
            data[idx] = arr.isel(xc=xcs, yc=ycs)

    return data


def process_regions(region: tuple,
                    data: tuple) -> tuple:
    """

    :param region:
    :param data:

    :return:
    """

    assert len(region) == 4, "Region needs to be a list of four integers"
    x1, y1, x2, y2 = region
    assert x2 > x1 and y2 > y1, "Region is not valid"

    for idx, arr in enumerate(data):
        if arr is not None:
            data[idx] = arr[..., (432 - y2):(432 - y1), x1:x2]
    return data


import copy
import fnmatch
import ftplib
import logging
import os

import datetime as dt
from ftplib import FTP

import dask
from distributed import Client, LocalCluster
import numpy as np
import pandas as pd
import xarray as xr

from icenet.data.cli import download_args
from icenet.data.producers import Downloader
from icenet.data.sic.mask import Masks
from icenet.utils import Hemisphere, run_command
from icenet.data.sic.utils import SIC_HEMI_STR

"""

"""

invalid_sic_days = {
    Hemisphere.NORTH: [
        *[d.date() for d in
          pd.date_range(dt.date(1979, 5, 21), dt.date(1979, 6, 4))],
        *[d.date() for d in
          pd.date_range(dt.date(1979, 6, 10), dt.date(1979, 6, 26))],
        dt.date(1979, 7, 1),
        *[d.date() for d in
          pd.date_range(dt.date(1979, 7, 24), dt.date(1979, 7, 28))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 1, 4), dt.date(1980, 1, 10))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 2, 27), dt.date(1980, 3, 4))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 3, 16), dt.date(1980, 3, 22))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 4, 9), dt.date(1980, 4, 15))],
        *[d.date() for d in
          pd.date_range(dt.date(1981, 2, 27), dt.date(1981, 3, 5))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 8, 12), dt.date(1984, 8, 24))],
        dt.date(1984, 9, 14),
        *[d.date() for d in
          pd.date_range(dt.date(1985, 9, 22), dt.date(1985, 9, 28))],
        *[d.date() for d in
          pd.date_range(dt.date(1986, 3, 29), dt.date(1986, 7, 1))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 1, 3), dt.date(1987, 1, 19))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 1, 29), dt.date(1987, 2, 2))],
        dt.date(1987, 2, 23),
        *[d.date() for d in
          pd.date_range(dt.date(1987, 2, 26), dt.date(1987, 3, 2))],
        dt.date(1987, 3, 13),
        *[d.date() for d in
          pd.date_range(dt.date(1987, 3, 22), dt.date(1987, 3, 26))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 4, 3), dt.date(1987, 4, 17))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 12, 1), dt.date(1988, 1, 12))],
        dt.date(1989, 1, 3),
        *[d.date() for d in
          pd.date_range(dt.date(1990, 12, 21), dt.date(1990, 12, 26))],
        dt.date(1979, 5, 28),
        dt.date(1979, 5, 30),
        dt.date(1979, 6, 1),
        dt.date(1979, 6, 3),
        dt.date(1979, 6, 11),
        dt.date(1979, 6, 13),
        dt.date(1979, 6, 15),
        dt.date(1979, 6, 17),
        dt.date(1979, 6, 19),
        dt.date(1979, 6, 21),
        dt.date(1979, 6, 23),
        dt.date(1979, 6, 25),
        dt.date(1979, 7, 1),
        dt.date(1979, 7, 25),
        dt.date(1979, 7, 27),
        dt.date(1984, 9, 14),
        dt.date(1987, 1, 16),
        dt.date(1987, 1, 18),
        dt.date(1987, 1, 30),
        dt.date(1987, 2, 1),
        dt.date(1987, 2, 23),
        dt.date(1987, 2, 27),
        dt.date(1987, 3, 1),
        dt.date(1987, 3, 13),
        dt.date(1987, 3, 23),
        dt.date(1987, 3, 25),
        dt.date(1987, 4, 4),
        dt.date(1987, 4, 6),
        dt.date(1987, 4, 10),
        dt.date(1987, 4, 12),
        dt.date(1987, 4, 14),
        dt.date(1987, 4, 16),
        dt.date(1987, 4, 4),
        dt.date(1990, 1, 26),
        dt.date(2022, 11, 9),
    ],
    Hemisphere.SOUTH: [
        dt.date(1979, 2, 5),
        dt.date(1979, 2, 25),
        dt.date(1979, 3, 23),
        *[d.date() for d in
          pd.date_range(dt.date(1979, 3, 26), dt.date(1979, 3, 30))],
        dt.date(1979, 4, 12),
        dt.date(1979, 5, 16),
        *[d.date() for d in
          pd.date_range(dt.date(1979, 5, 21), dt.date(1979, 5, 27))],
        *[d.date() for d in
          pd.date_range(dt.date(1979, 7, 10), dt.date(1979, 7, 18))],
        dt.date(1979, 8, 10),
        dt.date(1979, 9, 3),
        *[d.date() for d in
          pd.date_range(dt.date(1980, 1, 4), dt.date(1980, 1, 10))],
        dt.date(1980, 2, 16),
        *[d.date() for d in
          pd.date_range(dt.date(1980, 2, 27), dt.date(1980, 3, 4))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 3, 14), dt.date(1980, 3, 22))],
        dt.date(1980, 3, 31),
        *[d.date() for d in
          pd.date_range(dt.date(1980, 4, 9), dt.date(1980, 4, 15))],
        dt.date(1980, 4, 22),
        *[d.date() for d in
          pd.date_range(dt.date(1981, 2, 27), dt.date(1981, 3, 5))],
        dt.date(1981, 6, 10),
        *[d.date() for d in
          pd.date_range(dt.date(1981, 8, 3), dt.date(1982, 8, 9))],
        dt.date(1982, 8, 6),
        *[d.date() for d in
          pd.date_range(dt.date(1983, 7, 7), dt.date(1983, 7, 11))],
        dt.date(1983, 7, 22),
        dt.date(1984, 6, 12),
        *[d.date() for d in
          pd.date_range(dt.date(1984, 8, 12), dt.date(1984, 8, 24))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 9, 13), dt.date(1984, 9, 17))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 10, 3), dt.date(1984, 10, 9))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 11, 18), dt.date(1984, 11, 22))],
        dt.date(1985, 7, 23),
        *[d.date() for d in
          pd.date_range(dt.date(1985, 9, 22), dt.date(1985, 9, 28))],
        *[d.date() for d in
          pd.date_range(dt.date(1986, 3, 29), dt.date(1986, 11, 2))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 1, 3), dt.date(1987, 1, 15))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 12, 1), dt.date(1988, 1, 12))],
        dt.date(1990, 8, 14),
        dt.date(1990, 8, 15),
        dt.date(1990, 8, 24),
        *[d.date() for d in
          pd.date_range(dt.date(1990, 12, 22), dt.date(1990, 12, 26))],
        dt.date(1979, 2, 5),
        dt.date(1979, 2, 25),
        dt.date(1979, 3, 23),
        dt.date(1979, 3, 27),
        dt.date(1979, 3, 29),
        dt.date(1979, 4, 12),
        dt.date(1979, 5, 16),
        dt.date(1979, 7, 11),
        dt.date(1979, 7, 13),
        dt.date(1979, 7, 15),
        dt.date(1979, 7, 17),
        dt.date(1979, 8, 10),
        dt.date(1979, 9, 3),
        dt.date(1980, 2, 16),
        dt.date(1980, 3, 15),
        dt.date(1980, 3, 31),
        dt.date(1980, 4, 22),
        dt.date(1981, 6, 10),
        dt.date(1982, 8, 6),
        dt.date(1983, 7, 8),
        dt.date(1983, 7, 10),
        dt.date(1983, 7, 22),
        dt.date(1984, 6, 12),
        dt.date(1984, 9, 14),
        dt.date(1984, 9, 16),
        dt.date(1984, 10, 4),
        dt.date(1984, 10, 6),
        dt.date(1984, 10, 8),
        dt.date(1984, 11, 19),
        dt.date(1984, 11, 21),
        dt.date(1985, 7, 23),
        *[d.date() for d in
          pd.date_range(dt.date(1986, 7, 2), dt.date(1986, 11, 1))],
        dt.date(1990, 8, 14),
        dt.date(1990, 8, 15),
        dt.date(1990, 8, 24),
        dt.date(2022, 11, 9),
    ]
}

var_remove_list = ['time_bnds', 'raw_ice_conc_values', 'total_standard_error',
                   'smearing_standard_error', 'algorithm_standard_error',
                   'status_flag', 'Lambert_Azimuthal_Grid']


# This is adapted from the data/loaders implementations
class DaskWrapper:
    """

    :param dask_port:
    :param dask_timeouts:
    :param dask_tmp_dir:
    :param workers:
    """

    def __init__(self,
                 dask_port: int = 8888,
                 dask_timeouts: int = 60,
                 dask_tmp_dir: object = "/tmp",
                 workers: int = 8):

        self._dashboard_port = dask_port
        self._timeout = dask_timeouts
        self._tmp_dir = dask_tmp_dir
        self._workers = workers

    def dask_process(self,
                     *args,
                     method: callable,
                     **kwargs):
        """

        :param method:
        """
        dashboard = "localhost:{}".format(self._dashboard_port)

        with dask.config.set({
            "temporary_directory": self._tmp_dir,
            "distributed.comm.timeouts.connect": self._timeout,
            "distributed.comm.timeouts.tcp": self._timeout,
        }):
            cluster = LocalCluster(
                dashboard_address=dashboard,
                n_workers=self._workers,
                threads_per_worker=1,
                scheduler_port=0,
            )
            logging.info("Dashboard at {}".format(dashboard))

            with Client(cluster) as client:
                logging.info("Using dask client {}".format(client))
                ret = method(*args, **kwargs)
        return ret


class SICDownloader(Downloader):
    """Downloads OSI-SAF SIC data from 1979-present using OpenDAP.

    The dataset comprises OSI-450 (1979-2015) and OSI-430-b (2016-ownards)
    Monthly averages are-computed on the server-side.
    This script can take about an hour to run.

    The query URLs were obtained from the following sites:
        - OSI-450 (1979-2016): https://thredds.met.no/thredds/dodsC/osisaf/
            met.no/reprocessed/ice/conc_v2p0_nh_agg.html
        - OSI-430-b (2016-present): https://thredds.met.no/thredds/dodsC/osisaf/
            met.no/reprocessed/ice/conc_crb_nh_agg.html
        - OSI-430-a (2022-present): https://osi-saf.eumetsat.int/products/osi-430-as

    :param additional_invalid_dates:
    :param chunk_size:
    :param dates:
    :param delete_tempfiles:
    :param download:
    :param dtype:
    """
    def __init__(self,
                 *args,
                 additional_invalid_dates: object = (),
                 chunk_size: int = 10,
                 dates: object = (),
                 delete_tempfiles: bool = True,
                 download: bool = True,
                 dtype: object = np.float32,
                 parallel_opens: bool = True,
                 **kwargs):
        super().__init__(*args, identifier="osisaf", **kwargs)

        self._chunk_size = chunk_size
        self._dates = dates
        self._delete = delete_tempfiles
        self._download = download
        self._dtype = dtype
        self._parallel_opens = parallel_opens
        self._invalid_dates = invalid_sic_days[self.hemisphere] + \
            list(additional_invalid_dates)
        self._masks = Masks(north=self.north, south=self.south)

        self._ftp_osi450 = "/reprocessed/ice/conc/v2p0/{:04d}/{:02d}/"
        self._ftp_osi430b = "/reprocessed/ice/conc-cont-reproc/v2p0/{:04d}/{:02d}/"
        self._ftp_osi430a = "/reprocessed/ice/conc-cont-reproc/v3p0/{:04d}/{:02d}/"

        self._mask_dict = {
            month: self._masks.get_active_cell_mask(month)
            for month in np.arange(1, 12+1)
        }

    def download(self):
        """

        """
        hs = SIC_HEMI_STR[self.hemisphere_str[0]]
        data_files = []
        ftp = None
        var = "siconca"

        logging.info(
            "Not downloading SIC files, (re)processing NC files in "
            "existence already" if not self._download else
            "Downloading SIC datafiles to .temp intermediates...")

        cache = {}
        osi430b_start = dt.date(2016, 1, 1)
        osi430a_start = dt.date(2018, 11, 18)

        dt_arr = list(reversed(sorted(copy.copy(self._dates))))

        # Filtering dates based on existing data
        filter_years = sorted(set([d.year for d in dt_arr]))
        extant_paths = [
            os.path.join(self.get_data_var_folder(var),
                         "{}.nc".format(filter_ds))
            for filter_ds in filter_years
        ]
        extant_paths = [df for df in extant_paths if os.path.exists(df)]

        if len(extant_paths) > 0:
            extant_ds = xr.open_mfdataset(extant_paths)
            exclude_dates = pd.to_datetime(extant_ds.time.values)
            logging.info("Excluding {} dates already existing from {} dates "
                         "requested.".format(len(exclude_dates), len(dt_arr)))

            dt_arr = sorted(list(set(dt_arr).difference(exclude_dates)))
            dt_arr.reverse()

            # We won't hold onto an active dataset during network I/O
            extant_ds.close()

        # End filtering

        while len(dt_arr):
            el = dt_arr.pop()

            if el in self._invalid_dates:
                logging.warning("Date {} is in invalid list".format(el))
                continue

            date_str = el.strftime("%Y_%m_%d")
            temp_path = os.path.join(
                self.get_data_var_folder(var, append=[str(el.year)]),
                "{}.temp".format(date_str))
            nc_path = os.path.join(
                self.get_data_var_folder(var, append=[str(el.year)]),
                "{}.nc".format(date_str))

            if not self._download:
                if os.path.exists(nc_path):
                    reproc_path = os.path.join(
                        self.get_data_var_folder(var,
                                                 append=[str(el.year)]),
                        "{}.reproc.nc".format(date_str))

                    logging.debug("{} exists, becoming {}".
                                  format(nc_path, reproc_path))
                    os.rename(nc_path, reproc_path)
                    data_files.append(reproc_path)
                else:
                    if os.path.exists(temp_path):
                        logging.info("Using existing {}".format(temp_path))
                        data_files.append(temp_path)
                    else:
                        logging.debug("{} does not exist".format(nc_path))
                continue
            else:
                if not os.path.isdir(os.path.dirname(temp_path)):
                    os.makedirs(os.path.dirname(temp_path), exist_ok=True)

                if os.path.exists(temp_path) or os.path.exists(nc_path):
                    logging.debug("{} file exists, skipping".format(date_str))
                    if not os.path.exists(nc_path):
                        data_files.append(temp_path)
                    continue

                if not ftp:
                    logging.info("FTP opening")
                    ftp = FTP('osisaf.met.no')
                    ftp.login()

                chdir_path = self._ftp_osi450 \
                    if el < osi430b_start else self._ftp_osi430b \
                    if el < osi430a_start else self._ftp_osi430a

                chdir_path = chdir_path.format(el.year, el.month)

                try:
                    ftp.cwd(chdir_path)

                    if chdir_path not in cache:
                        cache[chdir_path] = ftp.nlst()

                    cache_match = "ice_conc_{}_ease*_{:04d}{:02d}{:02d}*.nc".\
                        format(hs, el.year, el.month, el.day)
                    ftp_files = [el for el in cache[chdir_path]
                                 if fnmatch.fnmatch(el, cache_match)]

                    if len(ftp_files) > 1:
                        raise ValueError("More than a single file found: {}".
                                         format(ftp_files))
                    elif not len(ftp_files):
                        logging.warning("File is not available: {}".
                                        format(cache_match))
                        continue
                except ftplib.error_perm:
                    logging.warning("FTP error, possibly missing month chdir "
                                    "for {}".format(date_str))
                    continue

                with open(temp_path, "wb") as fh:
                    ftp.retrbinary("RETR {}".format(ftp_files[0]), fh.write)

                logging.debug("Downloaded {}".format(temp_path))
                data_files.append(temp_path)

        if ftp:
            ftp.quit()

        logging.debug("Files being processed: {}".format(data_files))

        if len(data_files):
            ds = xr.open_mfdataset(data_files,
                                   combine="nested",
                                   concat_dim="time",
                                   data_vars=["ice_conc"],
                                   drop_variables=var_remove_list,
                                   engine="netcdf4",
                                   chunks=dict(time=self._chunk_size,),
                                   parallel=self._parallel_opens)

            logging.debug("Processing out extraneous data")

            ds = ds.drop_vars(var_remove_list, errors="ignore")
            da = ds.resample(time="1D").mean().ice_conc

            da = da.where(da < 9.9e+36, 0.)  # Missing values
            da /= 100.  # Convert from SIC % to fraction

            for coord in ['lat', 'lon']:
                if coord not in da.coords:
                    logging.warning("Adding {} vals to coords, as missing in "
                                    "this the combined dataset".format(coord))
                    da.coords[coord] = self._get_missing_coordinates(var,
                                                                     hs,
                                                                     coord)

            # In experimenting, I don't think this is actually required
            for month, mask in self._mask_dict.items():
                da.loc[dict(time=(da['time.month'] == month))].values[:, ~mask] = 0.

            for date in da.time.values:
                day_da = da.sel(time=slice(date, date))

                if np.sum(np.isnan(day_da.data)) > 0:
                    logging.warning("NaNs detected, adding to invalid "
                                    "list: {}".format(date))
                    self._invalid_dates.append(pd.to_datetime(date))

            var_folder = self.get_data_var_folder(var)
            group_by = "time.year"

            for year, year_da in da.groupby(group_by):
                req_date = pd.to_datetime(year_da.time.values[0])

                year_path = os.path.join(
                    var_folder, "{}.nc".format(getattr(req_date, "year")))
                old_year_path = os.path.join(
                    var_folder, "old.{}.nc".format(getattr(req_date, "year")))

                if os.path.exists(year_path):
                    logging.info("Existing file needs concatenating: {} -> {}".
                                 format(year_path, old_year_path))
                    os.rename(year_path, old_year_path)
                    old_da = xr.open_dataarray(old_year_path)
                    year_da = year_da.drop_sel(time=old_da.time,
                                               errors="ignore")
                    year_da = xr.concat([old_da, year_da],
                                        dim="time").sortby("time")
                    old_da.close()
                    os.unlink(old_year_path)

                logging.info("Saving {}".format(year_path))
                year_da.compute()
                year_da.to_netcdf(year_path)

        self.missing_dates()

        if self._delete:
            for fpath in data_files:
                os.unlink(fpath)

    def missing_dates(self):
        """

        :return:
        """
        filenames = set([os.path.join(
            self.get_data_var_folder("siconca"),
            "{}.nc".format(el.strftime("%Y")))
            for el in self._dates])
        filenames = [f for f in filenames if os.path.exists(f)]

        logging.info("Opening for interpolation: {}".format(filenames))
        ds = xr.open_mfdataset(filenames,
                               combine="nested",
                               concat_dim="time",
                               chunks=dict(time=self._chunk_size, ),
                               parallel=self._parallel_opens)
        return self._missing_dates(ds.ice_conc)

    def _missing_dates(self, da: object) -> object:
        """

        :param da:
        :return:
        """
        if pd.Timestamp(1979, 1, 2) in da.time.values \
                and dt.date(1979, 1, 1) in self._dates\
                and pd.Timestamp(1979, 1, 1) not in da.time.values:
            da_1979_01_01 = da.sel(
                time=[pd.Timestamp(1979, 1, 2)]).copy().assign_coords(
                {'time': [pd.Timestamp(1979, 1, 1)]})
            da = xr.concat([da, da_1979_01_01], dim='time')
            da = da.sortby('time')

        dates_obs = [pd.to_datetime(date).date() for date in da.time.values]
        dates_all = [pd.to_datetime(date).date() for date in
                     pd.date_range(min(self._dates), max(self._dates))]

        # Weirdly, we were getting future warnings for timestamps, but unsure
        # where from
        invalid_dates = [pd.to_datetime(d).date() for d in self._invalid_dates]
        missing_dates = [date for date in dates_all
                         if date not in dates_obs
                         or date in invalid_dates]

        logging.info("Processing {} missing dates".format(len(missing_dates)))

        missing_dates_path = os.path.join(
            self.get_data_var_folder("siconca"), "missing_days.csv")

        with open(missing_dates_path, "a") as fh:
            for date in missing_dates:
                # FIXME: slightly unusual format for Ymd dates
                fh.write(date.strftime("%Y,%m,%d\n"))

        logging.debug("Interpolating {} missing dates".
                      format(len(missing_dates)))

        for date in missing_dates:
            if pd.Timestamp(date) not in da.time.values:
                logging.info("Interpolating {}".format(date))
                da = xr.concat([da,
                                da.interp(time=pd.to_datetime(date))],
                               dim='time')

        logging.debug("Finished interpolation")

        da = da.sortby('time')
        da.data = np.array(da.data, dtype=self._dtype)

        for date in missing_dates:
            date_str = pd.to_datetime(date).strftime("%Y_%m_%d")
            fpath = os.path.join(
                self.get_data_var_folder(
                    "siconca", append=[str(pd.to_datetime(date).year)]),
                "missing.{}.nc".format(date_str))

            if not os.path.exists(fpath):
                day_da = da.sel(time=slice(date, date))
                mask = self._mask_dict[pd.to_datetime(date).month]

                day_da.data[0][~mask] = 0.

                logging.info("Writing missing date file {}".format(fpath))
                day_da.to_netcdf(fpath)

        return da

    def _get_missing_coordinates(self, var, hs, coord):
        """

        :param var:
        :param hs:
        :param coord:
        """
        missing_coord_file = os.path.join(
            self.get_data_var_folder(var), "missing_coord_data.nc")

        if not os.path.exists(missing_coord_file):
            ftp_source_path = self._ftp_osi450.format(2000, 1)

            retrieve_cmd_template_osi450 = \
                "wget -m -nH -nd -O {} " \
                "ftp://osisaf.met.no{}/{}"
            filename_osi450 = \
                "ice_conc_{}_ease2-250_cdr-v2p0_200001011200.nc".format(hs)

            run_command(retrieve_cmd_template_osi450.format(
                missing_coord_file, ftp_source_path, filename_osi450))
        else:
            logging.info("Coordinate path {} already exists".
                         format(missing_coord_file))

        ds = xr.open_dataset(missing_coord_file,
                             drop_variables=var_remove_list,
                             engine="netcdf4").load()
        try:
            coord_data = getattr(ds, coord)
        except AttributeError as e:
            logging.exception("{} does not exist in coord reference file {}".
                              format(coord, missing_coord_file))
            raise RuntimeError(e)
        return coord_data


def main():
    args = download_args(var_specs=False,
                         workers=True,
                         extra_args=[
                            (("-u", "--use-dask"),
                             dict(action="store_true", default=False)),
                            (("-c", "--sic-chunking-size"),
                             dict(type=int, default=10)),
                            (("-dt", "--dask-timeouts"),
                             dict(type=int, default=120)),
                            (("-dp", "--dask-port"),
                             dict(type=int, default=8888))
                         ])

    logging.info("OSASIF-SIC Data Downloading")
    sic = SICDownloader(
        chunk_size=args.sic_chunking_size,
        dates=[pd.to_datetime(date).date() for date in
               pd.date_range(args.start_date, args.end_date, freq="D")],
        delete_tempfiles=args.delete,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
        parallel_opens=args.parallel_opens,
    )
    if args.use_dask:
        logging.warning("Attempting to use dask client for SIC processing")
        dw = DaskWrapper(workers=args.workers)
        dw.dask_process(method=sic.download)
    else:
        sic.download()



"""Tests for icenet commands (entry points)"""

from importlib_metadata import entry_points
import pytest


icenet_entry_points = [
    ep for ep in entry_points(group="console_scripts")
    if ep.module.startswith('icenet')
]


def test_have_entry_points():
    """Check that there is at least one entry point (to stop the other
    tests passing vacuously if these are moved)
    """
    assert len(icenet_entry_points) > 0
    

@pytest.mark.parametrize("entry_point", icenet_entry_points)
def test_entry_point_exists(entry_point):
    """Check that the entry point ep can be loaded correctly
    (parameterized over all icenet entry points)
    """
    # Check that the entry point can be loaded without raising an
    # exception
    entry_point.load()



import argparse
import datetime as dt
import logging
import os

import cf_units
import iris
import pandas as pd
import rasterio

import iris.analysis

from icenet.process.utils import date_arg
from icenet.utils import setup_logging

from icenet.plotting.utils import broadcast_forecast, get_forecast_ds


def reproject_output(forecast_file: object,
                     proj_file: object,
                     save_file: object) -> object:
    """

    :param forecast_file:
    :param proj_file:
    :param save_file:
    """
    logging.info("Loading forecast {}".format(forecast_file))
    forecast_cube = iris.load_cube(forecast_file)

    logging.info("Projecting as per {}".format(proj_file))
    gp = iris.load_cube(proj_file)

    forecast_cube.coord('projection_y_coordinate').convert_units('meters')
    forecast_cube.coord('projection_x_coordinate').convert_units('meters')

    logging.info("Attempting to reproject and save to {}".
                 format(save_file))
    latlon_cube = forecast_cube.regrid(gp, iris.analysis.Linear())
    iris.save(latlon_cube, save_file)


@setup_logging
def broadcast_args() -> argparse.Namespace:
    """CLI arguments for broadcasting several forecasts linearly through time

    """
    ap = argparse.ArgumentParser()
    ap.add_argument("-o", "--output-target", dest="target", default=None)
    ap.add_argument("start_date", type=date_arg)
    ap.add_argument("end_date", type=date_arg)
    ap.add_argument("datafiles", nargs="+")
    args = ap.parse_args()
    return args


def broadcast_main():
    """CLI entry point for icenet_output_broadcast

    """
    args = broadcast_args()
    broadcast_forecast(args.start_date,
                       args.end_date,
                       args.datafiles,
                       args.target)


@setup_logging
def reproject_args() -> argparse.Namespace:
    """CLI args for reprojecting against another file

    """
    ap = argparse.ArgumentParser()
    ap.add_argument("forecast_file")
    ap.add_argument("proj_file")
    ap.add_argument("save_file")
    args = ap.parse_args()
    return args


def reproject_main():
    """CLI entry point for icenet_output_reproject

    """
    args = reproject_args()
    reproject_output(args.forecast_file, args.proj_file, args.save_file)


@setup_logging
def geotiff_args() -> argparse.Namespace:
    """CLI args for creating geotiffs

    """
    ap = argparse.ArgumentParser()
    ap.add_argument("-o", "--output-path", default=".")
    ap.add_argument("-s", "--stddev",
                    help="Plot the standard deviation from the ensemble",
                    action="store_true",
                    default=False)
    ap.add_argument("-v", "--verbose", default=False, action="store_true")
    ap.add_argument("forecast_file")
    ap.add_argument("forecast_date")
    ap.add_argument("leadtimes",
                    help="Leadtimes to output, multiple as CSV, range as n..n",
                    type=lambda s: [int(i) for i in
                                    list(s.split(",") if "," in s else
                                         range(int(s.split("..")[0]),
                                               int(s.split("..")[1]) + 1) if ".." in s else
                                         [s])])

    args = ap.parse_args()
    return args


def create_geotiff_output():
    """CLI entry point for icenet_output_geotiff

    """
    args = geotiff_args()

    if not os.path.isdir(args.output_path):
        logging.warning("No directory at: {}, creating".
                        format(args.output_path))
        os.makedirs(args.output_path)
    elif os.path.isfile(args.output_path):
        raise RuntimeError("{} should be a directory and not existent...".
                           format(args.output_path))

    ds = get_forecast_ds(args.forecast_file,
                         args.forecast_date,
                         stddev=args.stddev)
    ds = ds.isel(time=0).transpose(..., "yc", "xc")

    # The projection information set when we create NetCDF output compliant
    # with CF standards still has units as meters, but the attributes on the
    # variables is 1000 meters. It's easier to reset this for the GeoTIFF output
    # else you'll get scale errors that are a pain to fix in downstream
    x_meters = ds.xc * 1000
    y_meters = ds.yc * 1000
    x_attrs = ds.xc.attrs
    y_attrs = ds.yc.attrs

    ds = ds.assign_coords(xc=x_meters, yc=y_meters)
    ds['xc'].attrs = x_attrs
    ds['yc'].attrs = y_attrs
    ds['xc'].attrs['units'] = cf_units.Unit('meters')
    ds['yc'].attrs['units'] = cf_units.Unit('meters')

    if type(ds.rio.crs) != rasterio.crs.CRS:
        raise RuntimeError("Did not extract CRS via the coordinates, ds.rio.crs"
                           " is not of type rasterio.crs.CRS")

    leadtimes = args.leadtimes \
        if args.leadtimes is not None \
        else list(range(1, int(max(ds.leadtime.values)) + 1))

    forecast_name = "{}.{}".format(
        os.path.splitext(os.path.basename(args.forecast_file))[0],
        args.forecast_date)

    logging.info("Selecting and outputting files from {} for {}".
                 format(args.forecast_file, args.forecast_date))

    for leadtime in leadtimes:
        pred_da = ds.sel(leadtime=leadtime)

        output_filename = os.path.join(args.output_path, "{}.{}.{}tiff".format(
            forecast_name,
            (pd.to_datetime(args.forecast_date) + dt.timedelta(
                days=leadtime)).strftime("%Y-%m-%d"),
            "" if not args.stddev else "stddev."
        ))

        logging.debug("Outputting leadtime {} to {}".
                      format(leadtime, output_filename))
        pred_da.rio.to_raster(output_filename)
